# -*- coding: utf-8 -*-
"""
Defines the PNConnection class.
"""

from pineboolib.interfaces.isqlcursor import ISqlCursor
from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.core import settings, decorators
from pineboolib.core.utils import utils_base
from pineboolib.interfaces import iconnection
from pineboolib import application
from pineboolib.application.database.orm import alembic_tools

from typing import List, Optional, Any, Union, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.interfaces import isqlcursor, isqldriver, isession  # pragma: no cover
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover
    from sqlalchemy.engine import base, result  # type: ignore [import] # pragma: no cover

    from pineboolib.application.database import pnconnectionmanager  # pragma: no cover

LOGGER = utils_base.logging.get_logger(__name__)


class PNConnection(QtCore.QObject, iconnection.IConnection):
    """Wrapper for database cursors which are used to emulate FLSqlCursor."""

    def __init__(
        self,
        db_name: str,
        db_host: str = "",
        db_port: int = 0,
        db_user_name: str = "",
        db_password: str = "",
        driver_alias: str = "",
    ) -> None:
        """Database connection through a sql driver."""

        super().__init__()
        self.conn = None
        self._transaction_level = 0
        self._driver = None
        self._db_name = db_name
        self._session_legacy = None
        self._session_atomic = None

        conn_manager = application.PROJECT.conn_manager
        self._conn_manager = conn_manager
        if conn_manager is None:
            raise Exception("conn_manager is not Initialized!")

        if "main_conn" in conn_manager.connections_dict.keys():
            main_conn_ = conn_manager.connections_dict["main_conn"]
            if main_conn_._db_name == self._db_name and not db_host:
                db_host = main_conn_._db_host
                db_port = main_conn_._db_port
                db_user_name = main_conn_._db_user_name
                db_password = main_conn_._db_password
                driver_alias = main_conn_.driverAlias()

        self._db_host = db_host
        self._db_port = db_port
        self._db_user_name = db_user_name
        self._db_password = db_password

        self._driver_name = conn_manager._drivers_sql_manager.aliasToName(driver_alias)

        if application.USE_INTERACTIVE_GUI:
            self._interactive_gui = "Pineboo" if not utils_base.is_library() else "Pinebooapi"
        else:
            self._interactive_gui = ""

        self._last_active_cursor = None
        self._last_error = ""
        self._is_open = False

    def connManager(self) -> "pnconnectionmanager.PNConnectionManager":
        """Return connection manager."""
        return self._conn_manager

    def database(self) -> "iconnection.IConnection":
        """Return self."""
        return self

    def db(self) -> "iconnection.IConnection":
        """Return self."""

        return self

    def connectionName(self) -> str:
        """Get the current connection name for this cursor."""
        return self._name

    def connection(self) -> "base.Connection":
        """Return base connection."""

        return self.driver().connection()

    def isOpen(self) -> bool:
        """Indicate if a connection is open."""

        return self._is_open and self.driver().is_open()

    def tables(self, tables_type: Union[str, int] = "") -> List[str]:
        """Return a list of available tables in the database, according to a given filter."""

        types = ["", "Tables", "SystemTables", "Views"]

        if isinstance(tables_type, int):
            item = ""
            if tables_type < len(types):
                item = types[tables_type]
        else:
            item = tables_type

        return self.driver().tables(item)

    def DBName(self) -> str:
        """Return the database name."""
        return self.driver().DBName()

    def driver(self) -> "isqldriver.ISqlDriver":
        """Return the instance of the driver that is using the connection."""
        if self._driver is None:
            self._driver = self._conn_manager._drivers_sql_manager.driver()

        return self._driver

    def session(self, raise_error: bool = True) -> "isession.PinebooSession":
        """
        Sqlalchemy session.

        When using the ORM option this function returns the session for sqlAlchemy.
        """
        if self._name == "main_conn":
            raise Exception("main_conn no es valido para session")
        mng = self.connManager()

        returned_session = (
            self._session_legacy if self._session_atomic is None else self._session_atomic
        )

        if not mng.is_valid_session(returned_session, raise_error):
            if returned_session is not None:
                mng.remove_session(returned_session)

            returned_session = self.driver().session()

            if self._session_atomic is not None:
                # self._session_atomic = returned_session
                raise Exception("la transacción atómica no es válida")
            else:
                LOGGER.debug("Nueva sesión %s --> %s" % (self._name, returned_session))
                self._session_legacy = returned_session

        if not returned_session:
            raise ValueError("Invalid session!")

        return returned_session

    def engine(self) -> "base.Engine":
        """Sqlalchemy connection."""

        return self.driver().engine()

    def conectar(
        self,
        db_name: str,
        db_host: str = "",
        db_port: int = 0,
        db_user_name: str = "",
        db_password: str = "",
    ) -> Union["base.Connection", bool]:
        """Request a connection to the database."""

        self._db_name = db_name
        self._db_host = db_host
        self._db_port = db_port
        self._db_user_name = db_user_name
        self._db_password = db_password
        # if self._db_name:
        #    self.driver().alias_ = self.driverName() + ":" + self._name
        self.driver().db_ = self
        LOGGER.debug("")
        result: Union["base.Connection", bool] = self.driver().connect(
            db_name, db_host, db_port, db_user_name, db_password
        )
        LOGGER.debug(
            " NEW CONNECTION NAME: %s, HOST: %s, PORT: %s, DB NAME: %s, USER NAME: %s, STATUS: %s",
            self._name,
            db_host,
            db_port,
            db_name,
            db_user_name,
            "FAILURE" if not result else "ESTABLISHED",
        )

        return result

    def driverName(self) -> str:
        """Return sql driver name."""

        return self.driver().driverName()

    def driverAlias(self) -> str:
        """Return sql driver alias."""
        return self.driver().alias_

    def driverNameToDriverAlias(self, name: str) -> str:
        """Return the alias from the name of a sql driver."""

        return self._conn_manager._drivers_sql_manager.nameToAlias(name)

    def lastError(self) -> str:
        """Return the last error reported by the sql driver."""

        return self.driver().last_error()

    def host(self) -> str:
        """Return the name of the database host."""

        return self._db_host

    def port(self) -> int:
        """Return the port used by the database."""

        return self._db_port

    def user(self) -> str:
        """Return the user name used by the database."""

        return self._db_user_name

    def returnword(self) -> str:
        """Return the password used by the database."""

        return self._db_password

    @decorators.deprecated
    def password(self) -> str:
        """Return the password used by the database."""

        return self._db_password

    def setInteractiveGUI(self, gui_name: str) -> None:
        """Set if it is an interactive GUI."""
        if application.USE_INTERACTIVE_GUI:
            self._interactive_gui = gui_name

    def formatValue(self, type_: str, value: Any, upper: bool) -> Any:
        """Return a correctly formatted value to be assigned as a where filter."""

        return self.driver().formatValue(type_, value, upper)

    def formatValueLike(self, type_: str, value: Any, upper: bool) -> str:
        """Return a correctly formatted value to be assigned as a WHERE LIKE filter."""

        return self.driver().formatValueLike(type_, value, upper)

    def lastActiveCursor(self) -> Optional["ISqlCursor"]:
        """Return the last active cursor in the sql driver."""

        return self._last_active_cursor

    def doTransaction(self, cursor: "isqlcursor.ISqlCursor") -> bool:
        """Make a transaction or savePoint according to transaction level."""

        if settings.CONFIG.value("application/isDebuggerMode", False):
            text_ = (
                "Creando punto de salvaguarda %s:%s" % (self._name, self._transaction_level)
                if self._transaction_level
                else "Iniciando Transacción... %s:%s" % (self._name, self._transaction_level)
            )
            application.PROJECT.message_manager().send("status_help_msg", "send", [text_])

        # LOGGER.warning(
        #    "Creando transaccion/savePoint número:%s, cursor:%s, tabla:%s",
        #    self._transaction_level,
        #    cursor.curName(),
        #    cursor.table(),
        # )
        if not self.transaction():
            return False
        if not self._transaction_level:
            self._last_active_cursor = cursor
            application.PROJECT.aq_app.emitTransactionBegin(cursor)

        self._transaction_level += 1
        cursor.private_cursor._transactions_opened.append(self._transaction_level)
        return True

    def transactionLevel(self) -> int:
        """Indicate the level of transaction."""

        return self._transaction_level

    def doRollback(self, cur: "isqlcursor.ISqlCursor") -> bool:
        """Drop a transaction or savepoint depending on the transaction level."""

        cancel = False
        if (
            cur.modeAccess() in (cur.Insert, cur.Edit)
            and cur.isModifiedBuffer()
            and cur.private_cursor._ask_for_cancel_changes
        ):
            msg_box = getattr(application.PROJECT.DGI, "msgBoxQuestion", None)
            if msg_box is not None:
                res = msg_box(
                    "Todos los cambios se cancelarán.¿Está seguro?", None, "Cancelar Cambios"
                )

                if res is not None:
                    if res == QtWidgets.QMessageBox.StandardButton.No:
                        return False

            cancel = True

        if self._transaction_level:
            if cur.private_cursor._transactions_opened:
                trans = cur.private_cursor._transactions_opened.pop()
                if not trans == self._transaction_level:
                    LOGGER.warning(
                        "FLSqlDatabase: El cursor %s va a deshacer la transacción %s pero la última que inició es la %s",
                        cur.curName(),
                        self._transaction_level,
                        trans,
                    )
            else:
                LOGGER.warning(
                    "FLSqlDatabaser : El cursor va a deshacer la transacción %s pero no ha iniciado ninguna",
                    self._transaction_level,
                )

            self._transaction_level -= 1
        else:
            return True

        if self._transaction_level:
            text_ = "Restaurando punto de salvaguarda %s:%s..." % (
                self._name,
                self._transaction_level,
            )
        else:
            text_ = "Deshaciendo Transacción... %s:%s" % (self._name, self._transaction_level)

        application.PROJECT.message_manager().send("status_help_msg", "send", [text_])

        # LOGGER.warning(
        #    "Desaciendo transacción número:%s, cursor:%s", self._transaction_level, cur.curName()
        # )

        if not self.rollback():
            return False

        cur.setModeAccess(cur.Browse)

        if not self._transaction_level:
            self._last_active_cursor = None
            application.PROJECT.aq_app.emitTransactionRollback(cur)

            if cancel:
                cur.select()

        return True

    def interactiveGUI(self) -> str:
        """Return if it is an interactive GUI."""

        return self._interactive_gui

    def doCommit(self, cur: "isqlcursor.ISqlCursor", notify: bool = True) -> bool:
        """Approve changes to a transaction or a save point based on your transaction level."""

        if not notify:
            cur.autoCommit.emit()

        if self._transaction_level:
            if cur.private_cursor._transactions_opened:
                trans = cur.private_cursor._transactions_opened.pop()
                if not trans == self._transaction_level:
                    LOGGER.warning(
                        "El cursor %s va a terminar la transacción %s pero la última que inició es la %s",
                        cur.curName(),
                        self._transaction_level,
                        trans,
                        stack_info=True,
                    )
            else:
                LOGGER.warning(
                    "El cursor va a terminar la transacción %s pero no ha iniciado ninguna",
                    self._transaction_level,
                )

            self._transaction_level -= 1
        else:
            return True

        if self._transaction_level:
            text_ = "Liberando punto de salvaguarda %s:%s..." % (
                self._name,
                self._transaction_level,
            )
        else:
            text_ = "Terminando Transacción... %s:%s" % (self._name, self._transaction_level)

        application.PROJECT.message_manager().send("status_help_msg", "send", [text_])

        # LOGGER.warning(
        #    "Aceptando transacción número:%s, cursor:%s", self._transaction_level, cur.curName()
        # )

        if not self.commit():
            return False

        if not self._transaction_level:
            self._last_active_cursor = None
            application.PROJECT.aq_app.emitTransactionEnd(cur)

        if notify:
            cur.setModeAccess(cur.Browse)

        return True

    def canDetectLocks(self) -> bool:
        """Indicate if the connection detects locks in the database."""

        return self.driver().canDetectLocks()

    def canOverPartition(self) -> bool:
        """Return True if the database supports the OVER statement."""

        return self.connManager().dbAux().driver().canOverPartition()

    def Mr_Proper(self):
        """Clean the database of unnecessary tables and records."""

        self.connManager().dbAux().driver().Mr_Proper()

    def transaction(self) -> bool:
        """Create a transaction."""
        try:
            session_ = self.session()

            if not session_.in_transaction():  # type: ignore [attr-defined]
                LOGGER.debug(
                    "%s: ISOLATION LEVEL %s"
                    % (self._name, session_.connection().get_isolation_level())
                )
                session_.begin()
            else:
                session_.begin_nested()
            return True
        except Exception as error:
            self._last_error = "No se pudo crear la transacción: %s" % str(error)
            LOGGER.warning(self._last_error)

        return False

    def commit(self) -> bool:
        """Release a transaction."""

        try:
            session_ = self.session()
            trans_ = (
                session_.get_nested_transaction()  # type: ignore [attr-defined]
                if session_.in_nested_transaction()  # type: ignore [attr-defined]
                else session_.get_transaction()  # type: ignore [attr-defined]
            )

            trans_.commit()  # type: ignore [union-attr]
            return True
        except Exception as error:
            LOGGER.warning("Commit: %s", str(error), stack_info=True)
            self._last_error = "No se pudo aceptar la transacción: %s" % str(error)

        return False

    def rollback(self) -> bool:
        """Roll back a transaction."""

        try:
            session_ = self.session()
            trans_ = (
                session_.get_nested_transaction()  # type: ignore [attr-defined]
                if session_.in_nested_transaction()  # type: ignore [attr-defined]
                else session_.get_transaction()  # type: ignore [attr-defined]
            )
            trans_.rollback()  # type: ignore [union-attr]

            return True
        except Exception as error:
            self._last_error = "No se pudo deshacer la transacción: %s" % str(error)

        return False

    def nextSerialVal(self, table: str, field: str) -> int:
        """Indicate next available value of a serial type field."""

        return self.driver().nextSerialVal(table, field)

    def existsTable(self, name: str) -> bool:
        """Indicate the existence of a table in the database."""

        return self.driver().existsTable(name)

    def createTable(self, tmd: "pntablemetadata.PNTableMetaData") -> bool:
        """Create a table in the database, from a PNTableMetaData."""

        sql = self.driver().sqlCreateTable(tmd, True)
        if not sql:
            return False

        use_save_points = self.driver()._use_create_table_save_points

        if use_save_points:
            self.transaction()
        for single_sql in sql.split(";"):
            self.execute_query(single_sql)
            if self.driver().last_error():
                LOGGER.exception(
                    "createTable: Error happened executing sql: %s...%s"
                    % (single_sql[:80], str(self.driver().last_error()))
                )
                if use_save_points:
                    self.rollback()
                self.driver().set_last_error_null()
                return False

        if use_save_points:
            self.commit()
        return True

    def mismatchedTable(self, tablename: str, tmd: "pntablemetadata.PNTableMetaData") -> bool:
        """Compare an existing table with a PNTableMetaData and return if there are differences."""

        return self.connManager().dbAux().driver().mismatchedTable(tablename, tmd)

    def normalizeValue(self, text: str) -> Optional[str]:
        """Return the value of a correctly formatted string to the database type from a string."""
        return self.driver().normalizeValue(text)

    # def queryUpdate(self, name: str, update: str, filter: str) -> Optional[str]:
    #    """Return a correct UPDATE query for the database type."""

    #    return self.driver().queryUpdate(name, update, filter)

    def execute_query(self, qry) -> Optional["result.Result"]:  # type: ignore [name-defined]
        """Execute a query in a database cursor."""

        return self.driver().execute_query(qry)

    def alterTable(self, new_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Modify the fields of a table in the database based on the differences of two PNTableMetaData."""
        # print(
        #    "* ALTER TABLE %s * %s "
        #    % ("LEGACY" if application.USE_ALTER_TABLE_LEGACY else "ALEMBIC", new_metadata.name())
        # )
        if application.USE_ALTER_TABLE_LEGACY:
            return self.connManager().dbAux().driver().alterTable(new_metadata)

        alm = alembic_tools.Migration(self.connManager().dbAux())
        return alm.upgrade(new_metadata)

    def canRegenTables(self) -> bool:
        """Return if can regenerate tables."""

        return self.driver().canRegenTables() if self._driver_name else False

    def regenTable(self, table_name: str, mtd: "pntablemetadata.PNTableMetaData") -> bool:
        """Regenerate a table."""

        return self.driver().regenTable(table_name, mtd)

    def getTimeStamp(self) -> str:
        """Return timestamp."""

        return self.driver().getTimeStamp()

    def __str__(self):
        """Return the name of the database in text format."""

        return self.DBName()

    def __repr__(self):
        """Return the name of the database in text format."""

        return self.DBName()

    def close(self):
        """Close connection."""
        if self._session_legacy is not None:
            self._session_legacy.close()
            self._session_legacy = None
        self._is_open = False
        self.driver().close()

    def sqlLength(self, field_name: str, size: int) -> str:
        """Return length formated."""

        return self.driver().sqlLength(field_name, size)

    def resolve_dsn(self) -> str:
        """Return dsn data."""

        return self.driver().loadConnectionString(
            self._db_name, self._db_host, self._db_port, self._db_user_name, self._db_password
        )



"""Qgroupbox module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from pineboolib.core import decorators

from pineboolib.core import settings

from pineboolib import logging
from pineboolib.q3widgets import qwidget
from typing import Optional, Any

logger = logging.get_logger(__name__)


class QGroupBox(QtWidgets.QGroupBox, qwidget.QWidget):  # type: ignore [misc] # noqa: F821
    """QGroupBox class."""

    # style_str: str
    # _line_width: int
    presset = QtCore.pyqtSignal(int)
    selectedId: int
    line_width: int = 1

    def __init__(self, *args, **kwargs) -> None:
        """Inicialize."""
        if len(args):
            name = None
            parent = None
            if isinstance(args[0], str):
                name = args[0]
            else:
                parent = args[0]

            if len(args) > 1:
                if isinstance(args[1], str):
                    name = args[1]
                else:
                    parent = args[1]

            if parent is not None:
                super().__init__(parent, **kwargs)
            else:
                super().__init__(**kwargs)

            if name is not None:
                self.setObjectName(name)

        else:
            super().__init__()

        if not settings.CONFIG.value("ebcomportamiento/spacerLegacy", False):
            self.setSizePolicy(
                QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred
            )

        self.setContentsMargins(0, 2, 0, 2)

    def setLayout(self, layout: Optional["QtWidgets.QLayout"]) -> None:
        """Set layout to QGroupBox."""

        # layout.setContentsMargins(0, 0, 0, 0)
        # layout.setSpacing(0)
        super().setLayout(layout)  # type: ignore [arg-type]

    def setLineWidth(self, width: int) -> None:
        """Set line width."""

        style_ = "%s#%s {  border: %spx solid gray; margin-top: 20px; border-radius: 3px;}" % (
            type(self).__name__,
            self.objectName(),
            width,
        )
        self.line_width = width
        self.setStyleSheet(style_)

    def setTitle(self, title: Optional[str]) -> None:
        """Set title."""
        if self.line_width == 0:
            title = ""
        if title == "":
            self.setLineWidth(0)
        super().setTitle(title)  # type: ignore [arg-type]

    def get_enabled(self) -> bool:
        """Return if enabled."""
        return self.isEnabled()

    def set_enabled(self, value: bool) -> None:
        """Set enabled."""

        self.setDisabled(not value)

    @decorators.pyqt_slot(bool)
    def setShown(self, value: bool) -> None:
        """Set shown."""
        self.setVisible(value)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute especified by name."""

        if name == "title":
            self.setTitle(str(value))
        else:
            super().__setattr__(name, value)

    @decorators.not_implemented_warn
    def setFrameShadow(self, frame_shadow: None) -> None:
        """Set frame shadow."""

    @decorators.not_implemented_warn
    def setFrameShape(self, frame_shape: None) -> None:
        """Set frame shape."""

        pass

    @decorators.not_implemented_warn
    def newColumn(self) -> None:
        """Create a new column."""

        pass

    enabled = property(get_enabled, set_enabled)



"""
Singleton module.
"""
from typing import Any, Dict


class Singleton(type):
    """
    Singleton metaclass.

    Use this class as your metaclass to ensure only one instance of your class is created.
    """

    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Record class instance created to prevent duplicate references."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""FLareas_model module."""

from pineboolib.application.database.orm import basemodel

import sqlalchemy  # type: ignore [import]


class Flareas(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flareas class."""

    __tablename__ = "flareas"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flareas",
        "alias": "Áreas",
        "fields": [
            {
                "name": "bloqueo",
                "alias": "Bloqueo",
                "type": "unlock",
                "null": False,
                "default": True,
            },
            {
                "name": "idarea",
                "alias": "Área",
                "pk": True,
                "type": "string",
                "length": 15,
                "relations": [{"card": "1M", "table": "flmodules", "field": "idarea"}],
                "null": False,
            },
            {
                "name": "descripcion",
                "alias": "Descripción",
                "type": "string",
                "length": 100,
                "null": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    bloqueo = sqlalchemy.Column("bloqueo", sqlalchemy.Boolean)
    idarea = sqlalchemy.Column("idarea", sqlalchemy.String(15), primary_key=True)
    descripcion = sqlalchemy.Column("descripcion", sqlalchemy.String(100))

    # <--- Fields ---



"""Test_aqs module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestAQSql(unittest.TestCase):
    """TestAQSql Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_delete(self) -> None:
        """Delete test."""
        from pineboolib.application.database import pnsqlcursor
        from pineboolib.fllegacy.aqsobjects import aqsql

        cur_areas = pnsqlcursor.PNSqlCursor("flareas")
        cur_areas.setModeAccess(cur_areas.Insert)
        cur_areas.refreshBuffer()
        cur_areas.setValueBuffer("idarea", "X")
        cur_areas.setValueBuffer("descripcion", "descripcion area x")
        self.assertTrue(cur_areas.commitBuffer())
        cur_areas.select()
        self.assertEqual(cur_areas.size(), 1)
        aq_ = aqsql.AQSql()

        self.assertTrue(aq_.del_("flareas", "idarea='X'"))
        self.assertEqual(cur_areas.size(), 1)
        cur_areas.refresh()
        self.assertEqual(cur_areas.size(), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test classes module."""

from pineboolib.qsa import qsa
from PyQt6 import QtWidgets  # type: ignore[import]

import unittest
from pineboolib.loader.main import init_testing, finish_testing
import os


class TestClasses(unittest.TestCase):
    """Test classes."""

    _prueba = False

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_qdir(self) -> None:
        """Test qdir class."""
        # from pineboolib.core.settings import config

        # tmp_dir = config.value("ebcomportamiento/temp_dir")
        dir_ = qsa.QDir(".", "*.py *.pyo")
        # current_path = dir_.path()
        # dir_.setPath(tmp_dir)
        path_ = os.path.abspath(dir_.path())
        self.assertTrue(dir_.exists())
        self.assertTrue(dir_.isReadable())
        self.assertEqual(dir_.absPath(), qsa.QDir(path_).absPath())
        # dir_.setPath(current_path)

    def test_qtextstream(self) -> None:
        """Test qtextstream class."""

        from pineboolib import application

        txt_ = "Hola!"
        txt_2 = "Hola de nuevo!"
        file_1 = qsa.QFile("%s/test_qtextstream.txt" % application.PROJECT.tmpdir)
        self.assertTrue(file_1.open(qsa.File.WriteOnly | qsa.File.Append))

        text_stream = qsa.QTextStream()
        text_stream.setDevice(file_1.ioDevice())
        text_stream.opIn(txt_ + "\n")
        file_1.close()

        with open("%s/test_qtextstream.txt" % application.PROJECT.tmpdir) as file_3:
            read_data = file_3.read()
            self.assertEqual(read_data, "Hola!\n")

        file_2 = qsa.QFile("%s/test_qtextstream.txt" % application.PROJECT.tmpdir)
        self.assertTrue(file_2.open(qsa.File.WriteOnly | qsa.File.Append))

        text_stream = qsa.QTextStream()
        text_stream.setDevice(file_2.ioDevice())
        text_stream.opIn(txt_2 + "\n")
        file_2.close()

        with open("%s/test_qtextstream.txt" % application.PROJECT.tmpdir) as file_4:
            read_data = file_4.read()
            self.assertEqual(read_data, "Hola!\nHola de nuevo!\n")

    def test_qsproject(self) -> None:
        """Test qsproject."""

        value_1 = "flfactppal.iface.prueba"
        qsa.aqApp.setScriptEntryFunction("flfactppal.iface.prueba")

        value_2 = qsa.QSProject.entryFunction

        self.assertEqual(value_1, value_2)

    def test_aq_global_functions(self) -> None:
        """Test AQGlobal function."""

        from PyQt6 import QtWidgets

        qsa.sys.AQGlobalFunctions.set("saludo", self.saludo)
        btn = QtWidgets.QPushButton()
        qsa.sys.AQGlobalFunctions.mapConnect(btn, "clicked()", "saludo")
        self.assertFalse(self._prueba)
        btn.clicked.emit()  # type: ignore [attr-defined] # noqa: F821
        self.assertTrue(self._prueba)

    def test_sort(self) -> None:
        """Test array.sort function."""

        array_ = [1, 6, 3, 4, 2, 0, 0, 9]
        self.assertEqual(sorted(array_), [0, 0, 1, 2, 3, 4, 6, 9])
        self.assertEqual(qsa.Sort(self.function_sort).sort_(array_), [0, 1, 2, 3, 4, 6, 0, 9])

    def test_splice(self) -> None:
        """Test splice."""

        array_ = [1, 2, 3, 4, 5, 6]
        qsa.splice(array_, 3, 0, 8, 9)
        self.assertEqual(array_, [1, 2, 3, 8, 9, 4, 5, 6])
        qsa.splice(array_, 3, 2)
        self.assertEqual(array_, [1, 2, 3, 4, 5, 6])
        qsa.splice(array_, 3, 2, 9, 8)

    def test_object_class(self) -> None:
        """Test object class."""

        obj_ = qsa.ObjectClass()
        button = QtWidgets.QPushButton()
        obj_.module_connect(button, "clicked", self, "saluda")
        obj_.module_disconnect(button, "clicked", self, "saluda")

    def saludo(self) -> None:
        """AQGlobalFunction test."""
        self._prueba = True

    def function_sort(self, number_1: int, number_2: int) -> int:
        """Sorted function."""
        if number_1 == number_2:
            return 0
        elif number_1 > number_2:
            return 1
        else:
            return -1

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""Process Module."""

from PyQt6 import QtCore  # type: ignore

# from PyQt6.QtCore import pyqtSignal
import sys

# from pineboolib.core import decorators

from typing import Any, List, Optional, Iterable, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application import types  # noqa: F401 # pragma: no cover


class ProcessBaseClass(QtCore.QProcess):
    """Process base class."""

    _std_out: Optional[str]
    _std_error: Optional[str]
    _encoding: str

    def read_std_error(self) -> Any:
        """Return last std error."""

        return self._std_error

    def read_std_out(self) -> Any:
        """Return last std out."""
        return self._std_out

    def set_std_out(self, value: str) -> None:
        """Set last std out."""

        self._std_out = value

    def set_std_error(self, value: str) -> None:
        """Set last std error."""

        self._std_error = value

    def get_working_directory(self) -> str:
        """Return working directory."""

        return super().workingDirectory()

    def readLine(self) -> bytes:  # type: ignore [override] # noqa: F821
        """Readline overload."""

        return super().readLine().data()

    def set_working_directory(self, working_directory: str) -> None:
        """Set working directory."""

        super().setWorkingDirectory(working_directory)

    stdout: str = property(read_std_out, set_std_out)  # type: ignore [assignment] # noqa F821
    stderr: str = property(read_std_error, set_std_error)  # type: ignore [assignment] # noqa F821
    workingDirectory: str = property(  # type: ignore[assignment] # noqa : F821
        get_working_directory, set_working_directory
    )


class ProcessStatic(ProcessBaseClass):
    """Process static class."""

    @classmethod
    def executeNoSplit(cls, comando: Union[list, "types.Array"], stdin_buffer: str = "") -> int:
        """Execute command no splitted."""

        comando_ = []
        for item in comando:
            comando_.append(item)

        # programa = list_[0]
        # arguments = list_[1:]
        # self.setProgram(programa)
        # self.setArguments(arguments)
        process = Process()
        process.setProgram(comando_[0])
        argumentos = comando_[1:]
        process.setArguments(argumentos)

        process.start()

        stdin_as_bytes = stdin_buffer.encode(process._encoding)
        process.writeData(stdin_as_bytes)
        process.waitForFinished(30000)

        cls.stderr = process.readAllStandardError().data().decode(process._encoding)
        cls.stdout = process.readAllStandardOutput().data().decode(process._encoding)
        return process.exitCode()

    @classmethod
    def execute(
        cls, program: Union[str, List, "types.Array"], arguments: Optional[Iterable[str]] = None  # type: ignore [override]
    ) -> int:
        """Execute normal command."""
        comando_: List[str] = []
        if isinstance(program, list):
            comando_ = program
        else:
            comando_ = str(program).split(" ")

        process = Process()
        process.setProgram(comando_[0])
        argumentos = comando_[1:]
        process.setArguments(argumentos)

        process.start()
        process.waitForFinished(30000)
        cls.stderr = process.readAllStandardError().data().decode(process._encoding)
        cls.stdout = process.readAllStandardOutput().data().decode(process._encoding)
        return process.exitCode()


class Process(ProcessBaseClass):
    """Process class."""

    def __init__(self, *args) -> None:
        """Inicialize."""

        super().__init__()
        # cast(pyqtSignal, self.readyReadStandardOutput).connect(self.stdoutReady)
        # cast(pyqtSignal, self.readyReadStandardError).connect(self.stderrReady)
        self._encoding = sys.getfilesystemencoding()
        self.normalExit = self.ExitStatus.NormalExit  # pylint: disable=invalid-name
        self.crashExit = self.ExitStatus.CrashExit  # pylint: disable=invalid-name

        if args:
            self.setProgram(args[0])
            argumentos = args[1:]
            self.setArguments(argumentos)

    def stop(self) -> None:
        """Stop the process."""
        self.kill()

    def writeToStdin(self, stdin_) -> None:
        """Write data to stdin channel."""

        stdin_as_bytes = stdin_.encode(self._encoding)
        self.writeData(stdin_as_bytes)
        # self.closeWriteChannel()

    # @decorators.pyqtSlot()
    # def stdoutReady(self) -> None:
    #    self._stdout = str(self.readAllStandardOutput())

    # @decorators.pyqtSlot()
    # def stderrReady(self) -> None:
    #    self._stderr = str(self.readAllStandardError())

    def get_is_running(self) -> bool:
        """Return if the process is running."""

        return self.state() in (self.ProcessState.Running, self.ProcessState.Starting)

    def exitcode(self) -> Any:
        """Return exit code."""

        return self.exitCode()

    def readStderr(self) -> str:
        """Return standart error."""

        return self.stderr

    def readStdout(self) -> str:
        """Return standart output."""

        return self.stdout

    def executeNoSplit(self, comando: Union[list, "types.Array"], stdin_buffer: str = "") -> int:
        """Execute command no splitted."""

        comando_ = []
        for item in comando:
            comando_.append(item)

        # programa = list_[0]
        # arguments = list_[1:]
        # self.setProgram(programa)
        # self.setArguments(arguments)
        self.setProgram(comando_[0])
        argumentos = comando_[1:]
        self.setArguments(argumentos)

        self.start()

        stdin_as_bytes = stdin_buffer.encode(self._encoding)
        self.writeData(stdin_as_bytes)
        self.waitForFinished(30000)

        self.stderr = self.readAllStandardError().data().decode(self._encoding)
        self.stdout = self.readAllStandardOutput().data().decode(self._encoding)
        return self.exitCode()

    def execute(  # type: ignore[override] # noqa : F821
        self,
        program: Union[str, List[str], "types.Array"],
        arguments: Optional[Iterable[str]] = None,
    ) -> int:
        """Execute normal command."""
        comando_: List[str] = []
        if isinstance(program, list):
            comando_ = program
        else:
            comando_ = str(program).split(" ")

        self.setProgram(comando_[0])
        argumentos = comando_[1:]
        self.setArguments(argumentos)

        self.start()
        self.waitForFinished(30000)
        self.stderr = self.readAllStandardError().data().decode(self._encoding)
        self.stdout = self.readAllStandardOutput().data().decode(self._encoding)
        return self.exitCode()

    running = property(get_is_running)



"""Test require module."""

from pineboolib.application import qsadictmodules
from pineboolib.application.parsers import parser_qsa
from importlib import util
from pineboolib.qsa.tests import fixture_path
from pineboolib.loader.main import init_testing, finish_testing
import os

import unittest


class TestRequire(unittest.TestCase):
    """Test Require."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()
        parser_qsa.STRICT_MODE = True
        util_path = fixture_path("Import.py")
        spec = util.spec_from_file_location("Import", util_path)
        if spec and spec.loader is not None:
            module_instance = util.module_from_spec(spec)
            spec.loader.exec_module(module_instance)  # type: ignore [attr-defined]
            main_class = module_instance.FormInternalObj()  # type: ignore [attr-defined]
            qsadictmodules.QSADictModules.set_qsa_tree("formImport", main_class)  # type: ignore [arg-type]

    def test_basic_1(self) -> None:
        """Require test."""

        # print("*", dir(qsadictmodules.QSADictModules.qsa_dict_modules()))

        # req = qsa.require("test_require.qs")
        req = qsadictmodules.from_project("formImport").from_(
            os.path.join(os.path.dirname(__file__), "fixtures", "test_require.qs")
        )
        self.assertTrue(hasattr(req, "get"))

        instance_ = req(None)
        instance_.push("1")
        instance_.push("2")
        self.assertTrue(instance_.at(0), "1")
        self.assertTrue(instance_.at(1), "2")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""FlformSearchdb module."""

# -*- coding: utf-8 -*-
from pineboolib import logging, application

from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore[import]


from pineboolib.core import decorators, settings
from pineboolib.core.utils import utils_base

from pineboolib.application.database import pnsqlcursor

from pineboolib.fllegacy import flformdb

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.metadata import pnaction  # pragma: no cover


LOGGER = logging.get_logger(__name__)


class FLFormSearchDB(flformdb.FLFormDB):
    """
    Subclass of the FLFormDB class, designed to search for a record in a table.

    The behavior of choosing a record is modified for only
    close the form and so the object that invokes it can get
    of the cursor said record.

    It also adds OK and Cancel buttons. Accept indicates that it has been
    chosen the active record (same as double clicking on it or
    press the Enter key) and Cancel abort the operation.

    @author InfoSiAL S.L.
    """

    """
    Boton Aceptar
    """
    pushButtonAccept: Optional["QtWidgets.QToolButton"]

    """
    Almacena si se ha abierto el formulario con el método FLFormSearchDB::exec()
    """

    _accepting_rejecting: bool
    _in_exec: bool

    def __init__(self, *args) -> None:
        """
        Initialize.
        """
        action: "pnaction.PNAction"
        parent: Optional["QtWidgets.QWidget"] = None
        cursor: "pnsqlcursor.PNSqlCursor"

        if isinstance(args[0], str):
            action = application.PROJECT.conn_manager.manager().action(args[0])
            cursor = pnsqlcursor.PNSqlCursor(action.table())
            if len(args) > 1 and args[1]:
                parent = args[1]

        elif isinstance(args[1], str):
            action = application.PROJECT.conn_manager.manager().action(args[1])
            cursor = args[0]
            if len(args) > 2 and args[2]:
                parent = args[2]
        elif isinstance(args[0], pnsqlcursor.PNSqlCursor):
            cursor = args[0]
            parent = args[1]
            action_ = cursor.action()
            if action_:
                action = action_

        else:
            raise Exception("Wrong size of arguments")

        if not parent:
            parent = QtWidgets.QApplication.activeModalWidget()

        if cursor is None:
            LOGGER.warning("Se ha llamado a FLFormSearchDB sin nombre de action o cursor")
            return

        if application.PROJECT.conn_manager is None:
            raise Exception("Project is not connected yet")

        super().__init__(action, parent, load=False)

        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        if cursor.actionName() != action.name():
            cursor.setAction(action)

        self.setCursor(cursor)

        self.accepted_ = False
        self._in_exec = False
        self.loop = False
        self._accepting_rejecting = False
        self.pushButtonAccept = None  # pylint: disable=invalid-name

        self.load()
        self.initForm()
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

    def load(self):
        """Load control."""

        super().load()

        action = application.PROJECT.actions[self._action.name()]
        action.load_master_widget()
        if action._master_widget is not None and not utils_base.is_library():
            action._master_widget._form = self  # type: ignore [assignment] # noqa: F821

    def setAction(self, action: "pnaction.PNAction") -> None:
        """Set a action."""

        if self.cursor_:
            self.cursor_.setAction(action)

    """
    formReady = QtCore.pyqtSignal()
    """

    def loadControls(self) -> None:
        """Load form controls."""

        self.init_tool_bar()

        size_policy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy(0), QtWidgets.QSizePolicy.Policy(0)
        )
        size_policy.setHeightForWidth(True)

        push_button_size = self._icon_size
        if settings.CONFIG.value("application/isDebuggerMode", False):
            pushButtonExport = QtWidgets.QToolButton(self)  # pylint: disable=invalid-name
            pushButtonExport.setObjectName("pushButtonExport")
            pushButtonExport.setSizePolicy(size_policy)
            pushButtonExport.setMinimumSize(push_button_size)
            pushButtonExport.setMaximumSize(push_button_size)
            pushButtonExport.setIcon(
                QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-properties.png"))
            )
            pushButtonExport.setShortcut(QtGui.QKeySequence(self.tr("F3")))
            pushButtonExport.setWhatsThis("Exportar a XML(F3)")
            pushButtonExport.setToolTip("Exportar a XML(F3)")
            pushButtonExport.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.bottomToolbar.layout().addWidget(pushButtonExport)  # type: ignore [union-attr]
            pushButtonExport.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.exportToXml
            )

            if settings.CONFIG.value("ebcomportamiento/show_snaptshop_button", False):
                push_button_snapshot = QtWidgets.QToolButton(self)
                push_button_snapshot.setObjectName("pushButtonSnapshot")
                push_button_snapshot.setSizePolicy(size_policy)
                push_button_snapshot.setMinimumSize(push_button_size)
                push_button_snapshot.setMaximumSize(push_button_size)
                push_button_snapshot.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-paste.png"))
                )
                push_button_snapshot.setShortcut(QtGui.QKeySequence(self.tr("F8")))
                push_button_snapshot.setWhatsThis("Capturar pantalla(F8)")
                push_button_snapshot.setToolTip("Capturar pantalla(F8)")
                push_button_snapshot.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(push_button_snapshot)  # type: ignore [union-attr]
                push_button_snapshot.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.saveSnapShot
                )

            spacer = QtWidgets.QSpacerItem(
                20, 20, QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed
            )
            self.bottomToolbar.layout().addItem(spacer)  # type: ignore [union-attr]

        if not self.pushButtonAccept:
            self.pushButtonAccept = QtWidgets.QToolButton(self)
            self.pushButtonAccept.setObjectName("pushButtonAccept")
            self.pushButtonAccept.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.accept
            )

        self.pushButtonAccept.setSizePolicy(size_policy)
        self.pushButtonAccept.setMaximumSize(push_button_size)
        self.pushButtonAccept.setMinimumSize(push_button_size)
        self.pushButtonAccept.setIcon(
            QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-save.png"))
        )
        # pushButtonAccept->setAccel(QtGui.QKeySequence(Qt::Key_F10)); FIXME
        self.pushButtonAccept.setFocus()
        self.pushButtonAccept.setWhatsThis("Seleccionar registro actual y cerrar formulario (F10)")
        self.pushButtonAccept.setToolTip("Seleccionar registro actual y cerrar formulario (F10)")
        self.pushButtonAccept.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.bottomToolbar.layout().addWidget(self.pushButtonAccept)  # type: ignore [union-attr]
        self.pushButtonAccept.show()

        if not self.pushButtonCancel:
            self.pushButtonCancel = QtWidgets.QToolButton(self)
            self.pushButtonCancel.setObjectName("pushButtonCancel")
            self.pushButtonCancel.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.reject
            )

        self.pushButtonCancel.setSizePolicy(size_policy)
        self.pushButtonCancel.setMaximumSize(push_button_size)
        self.pushButtonCancel.setMinimumSize(push_button_size)
        self.pushButtonCancel.setIcon(
            QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-stop.png"))
        )
        self.pushButtonCancel.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        # pushButtonCancel->setAccel(Esc); FIXME
        self.pushButtonCancel.setWhatsThis("Cerrar formulario sin seleccionar registro (Esc)")
        self.pushButtonCancel.setToolTip("Cerrar formulario sin seleccionar registro (Esc)")
        self.bottomToolbar.layout().addWidget(self.pushButtonCancel)  # type: ignore [union-attr]
        self.pushButtonCancel.show()
        if self.cursor_ is None:
            raise Exception("Cursor is empty!.")
        self.cursor_.setEdition(False)
        self.cursor_.setBrowse(False)
        self.cursor_.recordChoosed.connect(self.accept)

    def exec_(self, valor: Optional[str] = None) -> bool:
        """
        Show the form and enter a new event loop to wait, to select record.

        The name of a cursor field is expected
        returning the value of that field if the form is accepted
        and a QVariant :: Invalid if canceled.

        @param valor Name of a form cursor field
        @return The value of the field if accepted, or False if canceled
        """

        if not self.cursor_:
            return False

        if self.cursor_.isLocked():
            self.cursor_.setModeAccess(pnsqlcursor.PNSqlCursor.Browse)

        if self.loop or self._in_exec:
            print("FLFormSearchDB::exec(): Se ha detectado una llamada recursiva")
            if self.isHidden():
                super().show()
            if self._init_focus_widget:
                self._init_focus_widget.setFocus()
            return False

        self._in_exec = True
        self._accepting_rejecting = False
        self.accepted_ = False

        super().show()
        if self._init_focus_widget:
            self._init_focus_widget.setFocus()

        if self.iface:
            try:
                QtCore.QTimer.singleShot(50, self.iface.init)
            except Exception:
                pass

        if not self._is_closing:
            QtCore.QTimer.singleShot(0, self.emitFormReady)

        self.loop = True
        if self.eventloop:
            self.eventloop.exec()
        self.loop = False
        self._in_exec = False

        if self.accepted_ and valor:
            return self.cursor_.valueBuffer(valor)
        else:
            self.close()
            return False

    def setFilter(self, filter: str) -> None:
        """Apply a filter to the cursor."""

        if not self.cursor_:
            return
        previous_filter = self.cursor_.mainFilter()

        new_filter = ""
        if not previous_filter:
            new_filter = filter
        elif not filter or previous_filter.find(filter) > -1:
            return
        else:
            new_filter = "%s AND %s" % (previous_filter, filter)

        self.cursor_.setMainFilter(new_filter)

    def formClassName(self) -> str:
        """Return the class name of the form at runtime."""

        return "FormSearchDB"

    def formName(self) -> str:
        """
        Return internal form name.
        """

        return "formSearch%s" % self._id_mdi

    def closeEvent(self, event: Optional["QtGui.QCloseEvent"]) -> None:
        """Capture event close."""

        self.frameGeometry()
        # if self.focusWidget():
        #    fdb = self.focusWidget().parentWidget()
        #    try:
        #        if fdb and fdb.autoComFrame_ and fdb.autoComFrame_.isvisible():
        #            fdb.autoComFrame_.hide()
        #            return
        #    except Exception:
        #        pass

        if self.cursor_ and self.pushButtonCancel:
            if not self.pushButtonCancel.isEnabled():
                return

            self._is_closing = True
            self.setCursor(None)
        else:
            self._is_closing = True

        if self.isHidden():
            # self.saveGeometry()
            # self.closed.emit()
            super().closeEvent(event)  # type: ignore [arg-type]
            # self.deleteLater()
        else:
            self.reject()

    @decorators.pyqt_slot()
    def callInitScript(self) -> None:
        """Call the "init" function of the "masterprocess" script associated with the form."""

        pass

    @decorators.pyqt_slot()
    def hide(self) -> None:
        """Redefined for convenience."""

        if self.loop:
            self.loop = False
            self.eventloop.exit()

        if self.isHidden():
            return

        super().hide()

    @decorators.pyqt_slot()
    def accept(self) -> None:
        """Activate pressing the accept button."""

        if self._accepting_rejecting:
            return
        self.frameGeometry()
        if self.cursor_:
            try:
                self.cursor_.recordChoosed.disconnect(self.accept)
            except Exception:
                pass
        self._accepting_rejecting = True
        self.accepted_ = True
        self.saveGeometry()
        self.hide()

        parent = self.parent()
        if isinstance(parent, QtWidgets.QMdiSubWindow):
            parent.hide()

    @decorators.pyqt_slot()
    def reject(self) -> None:
        """Activate pressing the accept button."""

        if self._accepting_rejecting:
            return
        self.frameGeometry()
        if self.cursor_:
            try:
                self.cursor_.recordChoosed.disconnect(self.accept)
            except Exception:
                pass
        self._accepting_rejecting = True
        self.hide()

    @decorators.pyqt_slot()
    def show(self) -> None:
        """Redefined for convenience."""
        self.exec_()


# ===============================================================================
#     def setMainWidget(self, w: QtWidgets.QWidget = None) -> None:
#         """
#         Set widget as the main form.
#         """
#
#         if not self.cursor_:
#             return
#
#         if w:
#             w.hide()
#             self.mainWidget_ = w
# ===============================================================================



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flgroups_model module."""

from pineboolib.application.database.orm import basemodel

import sqlalchemy  # type: ignore [import] # noqa: F821


class Flgroups(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flgroups class."""

    __tablename__ = "flgroups"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flgroups",
        "alias": "Grupos de Usuarios",
        "fields": [
            {
                "name": "idgroup",
                "alias": "Nombre",
                "pk": True,
                "type": "string",
                "length": 30,
                "relations": [
                    {"card": "1M", "table": "flusers", "field": "idgroup"},
                    {"card": "1M", "table": "flacs", "field": "idgroup"},
                ],
                "null": False,
            },
            {"name": "descripcion", "alias": "Descripción", "type": "string", "length": 100},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    idgroup = sqlalchemy.Column("idgroup", sqlalchemy.String(30), primary_key=True)
    descripcion = sqlalchemy.Column("descripcion", sqlalchemy.String(100))


# <--- Fields ---



"""AQSButtonGroup module."""

from pineboolib.core.utils import logging
from pineboolib.q3widgets import qbuttongroup

from typing import Any

LOGGER = logging.get_logger(__name__)


class AQSButtonGroup(qbuttongroup.QButtonGroup):
    """AQSButtonGroup class."""

    def __init__(self, parent: Any, name="", *args):
        """Initialize."""

        super().__init__(parent)
        if name:
            self.setObjectName(name)

        if args:
            LOGGER.warning("LOS ARGUMENTOS NO ESTAN SOPORTADOS FIXME")



"""Test sessions module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestSessions(unittest.TestCase):
    """TestSessions Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_selective_dirty(self) -> None:
        """Test basic."""

        cursor = qsa.FLSqlCursor("fltest")
        cursor.db().transaction()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("double_field", 0.1)
        cursor.setValueBuffer("string_field", "a")
        cursor.setValueBuffer("int_field", 1)
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("double_field", 0.2)
        cursor.setValueBuffer("string_field", "b")
        cursor.setValueBuffer("int_field", 2)
        self.assertTrue(cursor.commitBuffer())
        class_ = qsa.orm_("fltest")
        numero = qsa.util.sqlSelect("fltest", "min(id)", "1=1")

        obj1 = class_.get(numero)
        self.assertTrue(obj1)
        setattr(obj1, "_new_object", False)
        obj1._common_init()
        curobj = obj1.get_cursor()

        curobj.setValueBuffer("bool_field", True)
        self.assertEqual(curobj.valueBuffer("id"), numero)
        self.assertTrue(len(curobj._parent.changes()) > 0)

        self.assertTrue(
            len(curobj._parent.session.dirty) > 0,
            "dirty está vacio , y se esperaban datos (%s)" % (curobj._parent.session.dirty),
        )

        cursor2 = qsa.FLSqlCursor("fltest")
        numero = qsa.util.sqlSelect("fltest", "max(id)", "1=1")
        cursor2.select("id=%s" % (numero))
        self.assertTrue(cursor2.first())
        self.assertEqual(cursor2.valueBuffer("id"), numero)
        cursor2.setModeAccess(cursor2.Edit)
        cursor2.refreshBuffer()
        cursor2.setValueBuffer("bool_field", False)
        self.assertTrue(cursor2.commitBuffer())
        self.assertTrue(len(curobj._parent.session.dirty) == 1, curobj._parent.session.dirty)
        cursor3 = qsa.FLSqlCursor("fltest")
        cursor3.select("id=%s" % (numero))
        cursor3.first()
        cursor3.refreshBuffer()
        self.assertTrue(cursor3.valueBuffer("bool_field") is False)
        cursor.db().commit()

    def test_relation_session_flush(self):
        class_area = qsa.orm_("flareas", False)
        class_modulo = qsa.orm_("flmodules", False)

        self.assertFalse(class_area is None)
        self.assertFalse(class_modulo is None)

        obj_area = class_area()
        obj_area.bloqueo = False
        obj_area.idarea = "TS"
        obj_area.descripcion = "Area de pruebas de pendingRelationships"
        obj_area.save()

        session = obj_area._session

        obj_modulo_1 = class_modulo()
        obj_modulo_1.bloqueo = False
        obj_modulo_1.idmodulo = "M1"
        obj_modulo_1.idarea = obj_area.idarea
        obj_modulo_1.descripcion = "Modulo de pruebas 1 de pendingrelationships"
        obj_modulo_1.version = "0.1"
        obj_modulo_1.save()

        obj_modulo_2 = class_modulo()
        obj_modulo_2.bloqueo = False
        obj_modulo_2.idmodulo = "M2"
        obj_modulo_2.idarea = obj_area.idarea
        obj_modulo_2.descripcion = "Modulo de pruebas 2 de pendingrelationships"
        obj_modulo_2.version = "0.1"
        session.add(obj_modulo_2)

        self.assertTrue(obj_modulo_2 in session.new)
        self.assertTrue(len(session.dirty) == 0)

        obj_area.descripcion = "Area altered!"
        self.assertTrue(len(session.dirty) == 1)

        obj_modulo_1.descripcion = "."
        self.assertTrue(len(session.dirty) == 2)

        obj_modulo_2.save()

        self.assertTrue(len(session.dirty) == 2)

    def test_session_flush(self):
        class_area = qsa.orm_("flareas", False)

        self.assertFalse(class_area is None)

        obj_area_1 = class_area()
        session = obj_area_1._session
        obj_area_1.bloqueo = False
        obj_area_1.idarea = "TS11"
        obj_area_1.descripcion = "."
        obj_area_1.save()
        self.assertFalse(obj_area_1 in session.dirty)
        obj_area_1.descripcion = "1"
        self.assertTrue(obj_area_1 in session.dirty)
        obj_area_1.save()

        obj_area_2 = class_area()
        obj_area_2.bloqueo = False
        obj_area_2.idarea = "TS12"
        obj_area_2.descripcion = "."
        session.add(obj_area_2)
        self.assertTrue(obj_area_2 in session.new)
        obj_area_2.save()

        obj_area_3 = class_area()
        obj_area_3.bloqueo = False
        obj_area_3.idarea = "TS13"
        obj_area_3.descripcion = "."
        session.add(obj_area_3)

        obj_area_4 = class_area()
        obj_area_4.bloqueo = False
        obj_area_4.idarea = "TS14"
        obj_area_4.descripcion = "."
        session.add(obj_area_4)

        self.assertTrue(obj_area_3 in session.new)
        obj_area_1.descripcion = ".."
        obj_area_2.descripcion = ".."

        self.assertTrue(obj_area_1 in session.dirty)
        self.assertTrue(obj_area_2 in session.dirty)

        obj_area_3.save()
        self.assertTrue(obj_area_1 in session.dirty)
        self.assertTrue(obj_area_2 in session.dirty)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



# -*- coding: utf-8 -*-
"""
Module for PNSqlQuery class.
"""


from pineboolib.core.utils import logging

from pineboolib.application.utils import sql_tools
from pineboolib import application
from pineboolib.application import types

from typing import Union, List, Dict, Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces.ifieldmetadata import IFieldMetaData  # noqa: F401 # pragma: no cover
    from pineboolib.interfaces import iconnection  # pragma: no cover
    from pineboolib.interfaces import isqldriver
    from sqlalchemy.engine import result as result_engine  # type: ignore [import]
    from pineboolib.application.types import Array  # noqa: F401 # pragma: no cover
    from pineboolib.application.database import pngroupbyquery  # noqa: F401 # pragma: no cover
    from pineboolib.application.database import pnparameterquery  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class PNSqlQueryPrivate(object):
    """
    PNSqlQueryPrivate class.

    Store internal values of the query.
    """

    _name: str

    """
    Parte FROM de la consulta
    """
    _from: Optional[str]

    """
    Parte WHERE de la consulta
    """
    _where: Optional[str]

    """
    Parte ORDER BY de la consulta
    """
    _order_by: Optional[str]

    """
    Base de datos sobre la que trabaja
    """
    _db: "iconnection.IConnection"

    """
    Lista de parámetros
    """
    _parameter_dict: Dict[str, Any] = {}

    """
    Lista de grupos
    """
    _group_dict: Dict[int, str] = {}

    """
    Lista de nombres de los campos
    """
    _field_list: List[str]

    """
    Lista de nombres de las tablas que entran a formar
    parte en la consulta
    """
    _tables_list: List[str]

    _forward_only: bool
    _limit: Optional[int]
    _offset: Optional[int]

    def __init__(self, name: Optional[str] = None) -> None:
        """Create a new instance of PNSqlQueryPrivate."""

        if name:
            self._name = name
        self._parameter_dict = {}
        self._group_dict = {}
        self._tables_list = []
        self._field_list = []
        self._order_by = None
        self._where = None
        self._from = None

        self._forward_only = False
        self._limit = None
        self._offset = None


class PNSqlQuery(object):
    """
    Handle queries with specific features.

    It offers the functionality to handle QSqlQuery queries and also offers methods
    to work with parameterized queries and grouping levels.
    """

    _count_ref_query: int = 0
    _invalid_tables_list = False
    _is_active: bool
    _field_name_to_pos_dict: Optional[Dict[str, int]]
    _sql_inspector: "sql_tools.SqlInspector"
    _row: Optional["result_engine.Row"]  # type: ignore [name-defined]
    _datos: List["result_engine.Row"]  # type: ignore [name-defined]
    _posicion: int
    _last_query: str
    private_query: "PNSqlQueryPrivate"
    _data_size: int

    def __init__(
        self, cx=None, connection_name: Union[str, "iconnection.IConnection"] = "default"
    ) -> None:
        """
        Initialize a new query.
        """

        if application.PROJECT.conn_manager.mainConn() is None:
            raise Exception("Project is not connected yet")
        self._field_name_to_pos_dict = None
        self.private_query = PNSqlQueryPrivate(cx)
        self.private_query._db = (
            application.PROJECT.conn_manager.useConn(connection_name)
            if isinstance(connection_name, str)
            else connection_name
        )

        self.db().session()  # precarga.

        self._last_query = ""
        self._count_ref_query = self._count_ref_query + 1
        self._row = None
        self._datos = []
        self._invalid_tables_list = False
        self.private_query._field_list = []
        self._data_size = 0
        self._is_active = False

        retorno_qry = None
        if cx:
            retorno_qry = application.PROJECT.conn_manager.manager().query(cx, self)

        if retorno_qry:
            self.private_query = retorno_qry.private_query

    def __del__(self) -> None:
        """
        Delete cursor properties when closing.
        """

        try:
            # if self._connection is not None:
            #    self._connection.close()
            del self._sql_inspector
        except Exception:
            pass

    @property
    def sql_inspector(self) -> "sql_tools.SqlInspector":
        """
        Return a sql inspector instance.

        Collect a query and return information about its composition.
        @return sql_inspector
        """

        if not getattr(self, "_sql_inspector", None):
            self._sql_inspector = sql_tools.SqlInspector()

        return self._sql_inspector

    def exec_(self, sql: Optional[str] = "") -> bool:
        """
        Run a query.

        This can be specified or calculated from the values previously provided.
        @param sql. query text.
        @return True or False return if the execution is successful.
        """
        self._is_active = False

        sql = sql if sql else self.sql()

        if not sql:
            LOGGER.warning("exec_: no sql provided and PNSqlQuery.sql() also returned empty")
            return False

        self.sql_inspector.set_sql(sql)
        self.sql_inspector.resolve()
        if self.sql_inspector.suspected_injection():
            LOGGER.exception("Suspect sql injection : %s", self.sql_inspector._suspected_injection)
        if not self.isValid():
            LOGGER.error("exec_: invalid tables list found on query * %s *", sql)
            return False

        self._last_query = sql

        if self.private_query._db.driver()._parse_porc:
            sql = sql.replace("%", "%%")

        LOGGER.trace(
            "exec_: Ejecutando consulta: <%s> en <%s>", sql, self.db()._name
        )  # type: ignore [misc] # noqa: F821, F401
        result = self.db().execute_query(sql)
        try:
            self._datos = result.fetchall() if result and result.returns_rows else []  # type: ignore [attr-defined]
        except Exception as error:
            LOGGER.exception("ERROR SQLQUERY!: %s", str(error))
            self._datos = []

        self._posicion = -1

        if self.db().lastError():
            LOGGER.error("Error ejecutando consulta: <%s>\n%s", sql, self.db().lastError())
            self._invalid_tables_list = True

            # LOGGER.trace("Detalle:", stack_info=True)
            return False
        self._data_size = len(self._datos)
        self._is_active = True
        # conn.commit()
        LOGGER.trace("_exec: Rows: %s SQL: <%s>", self._data_size, sql)

        return True

    def addParameter(self, parameter: Optional["pnparameterquery.PNParameterQuery"]) -> None:
        """
        Add the parameter description to the parameter dictionary.

        @param p FLParameterQuery object with the description of the parameter to add.
        """

        if parameter is not None:
            self.private_query._parameter_dict[parameter.name()] = parameter.value()

    def addGroup(self, group: Optional["pngroupbyquery.PNGroupByQuery"]) -> None:
        """
        Add a group description to the group dictionary.

        @param g PNGroupByQuery object with the description of the group to add.
        """

        if group is not None:
            if not self.private_query._group_dict:
                self.private_query._group_dict = {}

            self.private_query._group_dict[group.level()] = group.field()

    def setName(self, name: str) -> None:
        """
        To set the name of the query.

        @param name. query name.
        """
        self.private_query._name = name

    def name(self) -> str:
        """
        To get the name of the query.
        """

        return self.private_query._name

    def select(self) -> str:
        """
        To get the SELECT part of the SQL statement from the query.

        @return text string with the query SELECT.
        """
        ret_: List[str] = (
            self.private_query._field_list
            if self.private_query._field_list
            else self.sql_inspector.field_names()
        )

        return ",".join(ret_)

    def from_(self) -> str:
        """
        To get the FROM part of the SQL statement from the query.

        @return text string with the query FROM.
        """

        return (
            self.private_query._from if self.private_query._from else self.sql_inspector.get_from()
        )

    def where(self) -> str:
        """
        To get the WHERE part of the SQL statement from the query.

        @return text string with the query WHERE.
        """

        return (
            self.private_query._where
            if self.private_query._where
            else self.sql_inspector.get_where()
        )

    def orderBy(self) -> Optional[str]:
        """
        To get the ORDERBY part of the SQL statement from the query.

        @return text string with the query ORDERBY.
        """

        return (
            self.private_query._order_by
            if self.private_query._order_by
            else self.sql_inspector.get_order_by()
        )

    def setSelect(self, select: Union[str, List, "Array"], sep: str = ",") -> None:
        """
        To set the SELECT part of the SQL statement of the query.

        @param s Text string with the SELECT part of the SQL statement that
            Generate the query. This string should NOT include the reserved word.
            SELECT, nor the character '*' as a wild card. Only support the list
            of fields that should appear in the query separated by the string
            indicated in the parameter 'sep'
        @param sep String used as a separator in the field list. Default the comma is used.
        """
        list_fields = []

        if isinstance(select, str):
            if sep in select:
                # s = s.replace(" ", "")

                prev = ""
                for child in select.split(sep):
                    field_ = prev + child
                    if field_.count("(") == field_.count(")"):
                        list_fields.append(field_)
                        prev = ""
                    else:
                        prev = "%s," % field_

        elif isinstance(select, list):
            list_fields = select
        else:
            list_fields = [value for key, value in select]

        self.private_query._field_list.clear()

        if not list_fields and isinstance(select, str) and not "*" == select:
            self.private_query._field_list.append(select)
        else:
            # fieldListAux = s.split(sep)
            # for f in s:
            #    f = str(f).strip()

            table: Optional[str]
            field: Optional[str]

            for child in list_fields:
                table = field = None
                try:
                    if child.startswith(" "):
                        child = child[1:]
                    table = child[: child.index(".")]
                    field = child[child.index(".") + 1 :]
                except Exception:
                    pass

                if field == "*" and table:
                    mtd = self.db().connManager().manager().metadata(table, True)
                    if mtd is not None:
                        self.private_query._field_list = [
                            "%s.%s as %s" % (table, field_name, field_name)
                            for field_name in mtd.fieldNames()
                        ]
                        if not mtd.inCache():
                            del mtd

                else:
                    self.private_query._field_list.append(child)

                # self.private_query.select_ = ",".join(self.private_query._field_list)

    def setFrom(self, from_: str) -> None:
        """
        To set the FROM part of the SQL statement of the query.

        @param f Text string with the FROM part of the SQL statement that generate the query
        """

        self.private_query._from = from_
        # self.private_query._from = f.strip_whitespace()
        # self.private_query._from = self.private_query._from.simplifyWhiteSpace()

    def setWhere(self, where_: str) -> None:
        """
        To set the WHERE part of the SQL statement of the query.

        @param s Text string with the WHERE part of the SQL statement that generates the query.
        """

        self.private_query._where = where_
        # self.private_query._where = w.strip_whitespace()
        # self.private_query._where = self.private_query._where.simplifyWhiteSpace()

    def setOrderBy(self, order_by: str) -> None:
        """
        To set the ORDER BY part of the SQL statement of the query.

        @param s Text string with the ORDER BY part of the SQL statement that generate the query
        """
        self.private_query._order_by = order_by
        # self.private_query._order_by = w.strip_whitespace()
        # self.private_query._order_by = self.private_query._order_by.simplifyWhiteSpace()

    def sql(self) -> str:
        """
        To get the full SQL statement of the query.

        This method joins the three parts of the query (SELECT, FROM AND WHERE),
        replace the parameters with the value they have in the dictionary and return all in a text string.
        @return Text string with the full SQL statement that generates the query.
        """
        # for tableName in self.private_query.tablesList_:
        #    if not self.private_query._db.manager().existsTable(tableName) and not self.private_query._db.manager().createTable(tableName):
        #        return

        res = None

        if not self.private_query._field_list:
            if self._last_query:
                return self._last_query

            LOGGER.warning("sql(): No select yet. Returning empty", stack_info=True)
            return ""

        select = ",".join(self.private_query._field_list)

        if not self.private_query._from:
            res = "SELECT %s" % select
        elif not self.private_query._where:
            res = "SELECT %s FROM %s" % (select, self.private_query._from)
        else:
            res = "SELECT %s FROM %s WHERE %s" % (
                select,
                self.private_query._from,
                self.private_query._where,
            )

        if self.private_query._group_dict and not self.private_query._order_by:
            res = res + " ORDER BY "
            res += ", ".join(
                [group_dict for key, group_dict in self.private_query._group_dict.items()]
            )

        elif self.private_query._order_by:
            res += " ORDER BY %s" % self.private_query._order_by

        if self.private_query._limit is not None:
            res += " LIMIT %s" % self.private_query._limit

        if self.private_query._offset is not None:
            if self.private_query._limit is None:
                res += " LIMIT %s" % 99999999
                LOGGER.warning("It is highly recommended to use limit next to offset")

            res += " OFFSET %s" % self.private_query._offset
            if self.private_query._order_by is None:
                LOGGER.warning("It is highly recommended to use order by next to offset")

        if self.private_query._parameter_dict:
            for key, parameter in self.private_query._parameter_dict.items():
                if parameter is None:
                    from PyQt6 import QtWidgets  # type: ignore[import]

                    dialog = QtWidgets.QInputDialog()

                    if dialog is not None:
                        parameter = dialog.getText(
                            QtWidgets.QApplication.activeWindow(),
                            "Entrada de parámetros de la consulta",
                            key,
                        )
                        if parameter:
                            parameter = parameter[0]

                res = res.replace(
                    "[%s]" % key, "'%s'" % parameter
                )  # FIXME: ajustar al tipo de dato pnparameterquery.setValue!!

        return res

    def parameterDict(self) -> Dict[str, "pnparameterquery.PNParameterQuery"]:
        """
         To obtain the parameters of the query.

        @return Parameter dictionary
        """
        return self.private_query._parameter_dict

    def groupDict(self) -> Dict[int, str]:
        """
        To obtain the grouping levels of the query.

        @return Dictionary of grouping levels.
        """

        return self.private_query._group_dict

    def fieldList(self, alternate_order: bool = False) -> List[str]:
        """
        To get the list of field names.

        @return List of text strings with the names of the fields in the query.
        """

        return (
            self.sql_inspector.field_names() or self.private_query._field_list
            if not alternate_order
            else self.private_query._field_list or self.sql_inspector.field_names()
        )

    def setGroupDict(self, groups_dict: Dict[int, str]) -> None:
        """
        Assign a parameter dictionary to the query parameter dictionary.

        The parameter dictionary of the FLGroupByQueryDict type, already built,
        It is assigned as the new group dictionary of the query, in the event that
        There is already a dictionary of groups, this is destroyed and overwritten by the new one.
        The dictionary passed to this method becomes the property of the query, and she is the
        responsible for deleting it. If the dictionary to be assigned is null or empty this
        method does nothing.

        @param gd Dictionary of parameters.
        """
        self.private_query._group_dict = groups_dict

    def setParameterDict(
        self, parameter_dict: Dict[str, "pnparameterquery.PNParameterQuery"]
    ) -> None:
        """
        Assign a group dictionary to the group dictionary of the query.

        The group dictionary of the FLParameterQueryDict type, already built,
        It is assigned as the new dictionary of query parameters, in the event that
        There is already a dictionary of parameters, it is destroyed and overwritten by the new one.
        The dictionary passed to this method becomes the property of the query, and she is the
        responsible for deleting it. If the dictionary to be assigned is null or empty this
        method does nothing.

        @param pd Parameter dictionary
        """

        self.private_query._parameter_dict = parameter_dict

    def showDebug(self) -> None:
        """
        Show the content of the query, by the standard output.

        It is intended only for debugging tasks.
        """
        if not self.isActive():
            LOGGER.warning(
                "DEBUG : La consulta no está activa : No se ha ejecutado exec() o la sentencia SQL no es válida"
            )

        LOGGER.warning("DEBUG : Nombre de la consulta : %s", self.private_query._name)
        LOGGER.warning("DEBUG : Niveles de agrupamiento :")
        if self.private_query._group_dict:
            for lev, field_name in self.private_query._group_dict.items():
                LOGGER.warning("**Nivel : %s", lev)
                LOGGER.warning("**Campo : %s", field_name)
        else:
            LOGGER.warning("**No hay niveles de agrupamiento")

        # LOGGER.warning("DEBUG : Parámetros : ")
        # if self.private_query._parameter_dict:
        #     if par in self.private_query._parameter_dict:
        #         LOGGER.warning("**Nombre : %s", par.name())
        #         LOGGER.warning("Alias : %s", par.alias())
        #         LOGGER.warning("Tipo : %s", par.type())
        #         LOGGER.warning("Valor : %s", par.value())
        # else:
        #     LOGGER.warning("**No hay parametros")

        LOGGER.warning("DEBUG : Sentencia SQL")
        LOGGER.warning("%s", self.sql())
        if not self.private_query._field_list:
            LOGGER.warning("DEBUG ERROR : No hay campos en la consulta")
            return

        linea = ""
        LOGGER.warning("DEBUG: Campos de la consulta : ")
        for field in self.private_query._field_list:
            LOGGER.warning("**%s", field)
            linea += "__%s" % self.value(field)
        LOGGER.warning("DEBUG : Contenido de la consulta : ")
        LOGGER.warning(linea)

    def value(self, field_name_or_pos: Union[str, int, None], raw: bool = False) -> Any:
        """
        Get the value of a query field.

        Given a name of a query field, this method returns a QVariant object
        with the value of that field. The name must correspond to the one placed in
        the SELECT part of the SQL statement of the query.

        @param n Name of the query field
        @param raw If TRUE and the value of the field is a reference to a large value
             (see FLManager :: storeLargeValue ()) returns the value of that reference,
             instead of content to which that reference points

        """

        if field_name_or_pos is None:
            LOGGER.trace("value::invalid use with n=None.", stack_info=True)
            return None

        pos: int = (
            self.sql_inspector.fieldNameToPos(field_name_or_pos.lower())
            if isinstance(field_name_or_pos, str)
            else int(field_name_or_pos)
        )

        try:
            ret = self._row[pos] if self._row else None  # type: ignore [index]
            return (
                self.sql_inspector.resolve_empty_value(pos)
                if ret in (None, "None")
                else self.sql_inspector.resolve_value(pos, ret, raw)
            )
        except Exception:
            LOGGER.exception("value::error retrieving row position %s", pos)

    def isNull(self, field_name: str) -> bool:
        """
        Indicate whether a query field is null or not.

        Given a name of a query field, this method returns true if the query field is null.
        The name must correspond to the one placed in
        the SELECT part of the SQL statement of the query.

        @param n Name of the query field
        """
        if not self._row:
            return True

        if isinstance(field_name, str):
            pos_ = self.fieldNameToPos(field_name)

            return self._row[pos_] in (None, "None")  # type: ignore [index]

        raise Exception("isNull. field not found %s" % field_name)

    def posToFieldName(self, position: int) -> str:
        """
        Return the field name, given its position in the query.

        @param p Position of the field in the query, start at zero and left to right.
        @return Name of the corresponding field. If the field does not exist, it returns None.
        """
        if self.sql_inspector.sql() == "":
            self.sql_inspector.set_sql(self.sql())
            self.sql_inspector.resolve()

        return self.sql_inspector.posToFieldName(position)
        # if p < 0 or p >= len(self.private_query._field_list):
        #    return None
        # ret_ = None
        # try:
        #    ret_ = self.private_query._field_list[p]
        # except Exception:
        #    pass

        # return ret_

    def fieldNameToPos(self, field_name: str) -> int:
        """
        Return the position of a field in the query, given its name.

        @param n Field Name.
        @return Position of the field in the query. If the field does not exist, return -1.
        """
        if self.sql_inspector.sql() == "":
            self.sql_inspector.set_sql(self.sql())
            self.sql_inspector.resolve()

        return self.sql_inspector.fieldNameToPos(field_name.lower())
        # i = 0
        # for field in self.private_query._field_list:
        #    if field.lower() == n.lower():
        #        return i
        #    i = i + 1
        # if n in self.private_query._field_list:
        #    return self.private_query._field_list.index(n)
        # else:
        #    return False

    def tablesList(self) -> List[str]:
        """
        To get the list of names of the query tables.

        @return List of names of the tables that become part of the query.
        """

        return (
            self.private_query._tables_list
            if self.private_query._tables_list
            else self.sql_inspector.table_names()
        )

    def setTablesList(self, table_list: Union[str, List, types.Array]) -> None:
        """
        Set the list of names of the query tables.

        @param table_list Text list (or a list) with the names of the tables separated by commas, e.g. "table1, table2, table3"
        """

        self.private_query._tables_list = []

        table_list = ",".join(table_list) if isinstance(table_list, list) else str(table_list)
        table_list = table_list.replace(" ", "")
        mng = self.db().connManager().manager()
        for tabla in table_list.split(","):
            if not mng.existsTable(tabla) and not mng.metadata(tabla):
                self._invalid_tables_list = True
                LOGGER.warning("setTablesList: table not found %r. Query will not execute.", tabla)
            self.private_query._tables_list.append(tabla)

    def setValueParam(self, param_name: str, value: Any) -> None:
        """
        Set the value of a parameter.

        @param name Parameter name.
        @param v Value for the parameters.
        """

        self.private_query._parameter_dict[param_name] = value

    def valueParam(self, param_name: str) -> Optional[Any]:
        """
        Get the value of a parameter.

        @param name Parameter name.
        """

        return (
            self.private_query._parameter_dict[param_name]
            if param_name in self.private_query._parameter_dict.keys()
            else None
        )

    def size(self) -> int:
        """
        Report the number of results returned by the query.

        @return number of results.
        """
        return self._data_size

    def fieldMetaDataList(self) -> List["IFieldMetaData"]:
        """
        To get the list of query field definitions.

        @return Object with the list of deficiencies in the query fields.
        """
        list_: List["IFieldMetaData"] = []
        dictado_ = self.sql_inspector.mtd_fields()
        for k in dictado_.keys():
            list_.append(dictado_[k])

        return list_

    def db(self) -> "iconnection.IConnection":
        """
        Get the database you work on.

        @return PNConnection user by the query.
        """
        return self.private_query._db

    def isValid(self) -> bool:
        """
        Return if the query has an invalid defined table.

        @return True or False.
        """
        value_inspector = True
        if self.sql_inspector is not None:
            if self.sql_inspector._invalid_tables:
                real_tables = self.db().tables()
                for table_name in self.sql_inspector._invalid_tables:
                    if table_name and table_name not in real_tables:
                        value_inspector = False
                        break

        value_invalid_tables_list = not self._invalid_tables_list
        return value_invalid_tables_list and value_inspector

    def isActive(self) -> bool:
        """
        Indicate whether the data has been collected completely.

        @return True or False.
        """
        return self._is_active

    def at(self) -> int:
        """
        Return the current position in the result list.

        @return line position.
        """

        return self._posicion

    def lastQuery(self) -> str:
        """
        Return the last query made.

        @return query string.
        """

        return self._last_query

    def numRowsAffected(self) -> int:
        """
        Return Number of lines selected in the query.

        @return number of lines.
        """
        return self._data_size

    def lastError(self) -> str:
        """Return last error if exists , empty elsewhere."""

        return self.db().lastError()

    def driver(self) -> "isqldriver.ISqlDriver":
        """Return sql driver."""

        return self.db().driver()

    def isForwardOnly(self) -> bool:
        """Return if is forward only enabled."""
        return self.private_query._forward_only

    def setForwardOnly(self, forward: bool) -> None:
        """Set forward only option value."""
        self.private_query._forward_only = forward

    def seek(self, position: int, relative=False) -> bool:
        """
        Position the cursor on a given result.

        @param i Position to search.
        @param relative Boolean indicates if the position is relative or absolut.
        @return True or False.
        """

        position += self._posicion if relative else 0

        if self._datos:
            if position >= 0 and position < self._data_size:
                self._posicion = position
                self._row = self._datos[self._posicion]
                return True

        return False

    def next(self) -> bool:
        """
        Position the query cursor in the next record.

        @return True or False.
        """

        if self._datos:
            self._posicion += 1
            if self._posicion < self._data_size:
                self._row = self._datos[self._posicion]
                return True

        return False

    def prev(self) -> bool:
        """
        Position the query cursor in the provious record.

        @return True or False.
        """

        if self._datos:
            self._posicion -= 1
            if self._posicion > -1:
                self._row = self._datos[self._posicion]
                return True

        return False

    def first(self) -> bool:
        """
        Position the query cursor in the first record.

        @return True or False.
        """

        if self._datos:
            self._posicion = 0
            self._row = self._datos[self._posicion]
            return True

        return False

    def last(self) -> bool:
        """
        Position the query cursor in the last record.

        @return True or False.
        """

        if self._datos:
            self._posicion = self._data_size - 1
            self._row = self._datos[self._posicion]
            return True

        return False

    def setLimit(self, limit: int) -> None:
        """Set limit."""

        self.private_query._limit = limit

    def setOffset(self, offset: int) -> None:
        """Set offset."""

        self.private_query._offset = offset



# -*- coding: utf-8 -*-
"""PNApplication Module."""


from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.core import decorators, settings
from pineboolib.core.utils import logging, utils_base

from pineboolib import application
from pineboolib.application.database import DB_SIGNALS, utils
from pineboolib.application.qsatypes import sysbasetype
from pineboolib.application import qsadictmodules
from pineboolib.application.utils import external

import sys
import os
from typing import Any, Optional, List, TextIO, cast, Union, Dict, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.application.database import (
        pnsqlcursor,
        pnsqlquery,
    )  # noqa: F401 # pragma: no cover
    from pineboolib.application.acls import pnaccesscontrollists
    from pineboolib.application.translator import pntranslator

    from pineboolib.interfaces import isqlcursor  # noqa: F401 # pragma: no cover
    from PyQt6 import QtXml, QtGui  # noqa: F401 # pragma: no cover
    from pineboolib.application import module as app_module  # pragma: no cover
    from pineboolib.interfaces import dgi_schema  # pragma: no cover
    from pineboolib.application.database import pnconnectionmanager  # pragma: no cover
    from pineboolib.fllegacy import flformdb  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FLPopupWarn(QtCore.QObject):
    """FLPoppupWarn Class."""

    # FIXME: Incomplete class!
    def __init__(self, mainwindow) -> None:
        """Inicialize."""

        self._main_window = mainwindow


class PNApplication(QtCore.QObject):
    """PNApplication Class."""

    _inicializing: bool
    _destroying: bool
    _ted_output: Optional["QtWidgets.QWidget"]
    _not_exit: bool
    _multi_lang_enabled: bool
    _multi_lang_id: str
    _translator: List["pntranslator.PNTranslator"]

    container_: Optional["QtWidgets.QWidget"]  # Contenedor actual??
    form_alone_: bool
    acl_: Optional["pnaccesscontrollists.PNAccessControlLists"]

    op_check_update_: bool
    style: bool
    flLargeMode: Optional[bool]

    init_single_fl_large: bool
    show_debug_: bool
    timer_idle_: Optional["QtCore.QTimer"]
    time_user_: "QtCore.QTimer"
    script_entry_function_: str
    _event_loop: Optional["QtCore.QEventLoop"]
    window_menu: Optional["QtWidgets.QMenu"] = None
    modules_menu: Optional["QtWidgets.QMenu"] = None

    transactionBegin: "QtCore.pyqtSignal" = QtCore.pyqtSignal()
    transactionEnd: "QtCore.pyqtSignal" = QtCore.pyqtSignal()
    transactionRollback: "QtCore.pyqtSignal" = QtCore.pyqtSignal()

    def __init__(self) -> None:
        """Create new FLApplication."""
        super().__init__()

        # self.project_ = None
        self.wb_ = None

        self._translator = []

        self.form_alone_ = False
        self._not_exit = False
        self.timer_idle_ = None
        # self.popup_warn_ = None
        self._inicializing = False
        self._destroying = False
        # self.fl_factory_ = None
        self.op_check_update_ = False
        self.window_menu = None
        DB_SIGNALS.notify_begin_transaction_ = False
        DB_SIGNALS.notify_end_transaction_ = False
        DB_SIGNALS.notify_roll_back_transaction_ = False
        self._ted_output = None
        self.style = False
        self.init_single_fl_large = False
        self.show_debug_ = True  # FIXME
        self.script_entry_function_ = ""
        self.flLargeMode = None

        self.acl_ = None
        # self.fl_factory_ = FLObjectFactory() # FIXME para un futuro
        # self.time_user_ = QtCore.QDateTime.currentDateTime() # Moved to pncontrolsfacotry.SysType
        self._multi_lang_enabled = False
        self._multi_lang_id = QtCore.QLocale().name()[:2].upper()

        self.locale_system_ = QtCore.QLocale.system()
        value = 1.1
        self.comma_separator = self.locale_system_.toString(value, "f", 1)[1]
        self.setObjectName("aqApp")
        self._event_loop = None

    @property
    def event_loop(self) -> "QtCore.QEventLoop":
        """Get Eventloop, create one if it does not exist."""
        if self._event_loop is None:
            self._event_loop = QtCore.QEventLoop()
        return self._event_loop

    def eventLoop(self) -> "QtCore.QEventLoop":
        """Create main event loop."""
        return QtCore.QEventLoop()

    def toXmlReportData(self, qry: "pnsqlquery.PNSqlQuery") -> "QtXml.QDomDocument":
        """Return xml from a query."""
        from pineboolib.fllegacy import flreportengine

        rpt_ = flreportengine.FLReportEngine()
        rpt_.setReportData(qry)
        ret = rpt_.reportData()
        return ret

    @decorators.not_implemented_warn
    def checkForUpdate(self):
        """Not used in Pineboo."""
        pass

    @decorators.not_implemented_warn
    def checkForUpdateFinish(self, option):
        """Not used in pineboo."""
        pass

    @decorators.not_implemented_warn
    def initfcgi(self):
        """Init for fast cgi."""
        pass

    @decorators.not_implemented_warn
    def addObjectFactory(self, new_object_factory):
        """Add object onctructor. unused."""
        pass

    @decorators.not_implemented_warn
    def callfcgi(self, call_function, argument_list):
        """Perform fastcgi call."""
        pass

    @decorators.not_implemented_warn
    def endfcgi(self):
        """End fastcgi call signal."""
        pass

    def localeSystem(self) -> "QtCore.QLocale":
        """Return locale of the system."""
        return self.locale_system_

    @decorators.not_implemented_warn
    def openQSWorkbench(self):
        """Open debugger. Unused."""
        pass

    def setMainWidget(self, main_widget) -> None:
        """Set mainWidget."""
        if application.PROJECT.main_window is not None:
            application.PROJECT.main_window.main_widget = main_widget
            if main_widget is not None:
                application.PROJECT.app.setActiveWindow(main_widget)

    @decorators.not_implemented_warn
    def makeStyle(self, style_):
        """Apply specified style."""
        pass

    def chooseFont(self) -> None:
        """Open font selector."""

        font_ = QtWidgets.QFontDialog().getFont()  # type: ignore[misc] # noqa: F821
        if font_:
            application.PROJECT.app.setFont(font_[0])
            save_ = [font_[0].family(), font_[0].pointSize(), font_[0].weight(), font_[0].italic()]

            settings.CONFIG.set_value("application/font", save_)

    def showStyles(self) -> None:
        """Open style selector."""
        if not self.style:
            self.initStyles()
        # if self.style:
        #    self.style.exec()

    @decorators.not_implemented_warn
    def showToggleBars(self):
        """Show toggle bars."""
        pass

    def setStyle(self, style_: str) -> None:
        """Change application style."""
        settings.CONFIG.set_value("application/style", style_)
        application.PROJECT.app.setStyle(style_)  # type: ignore [misc,call-overload] # noqa: F821

    def initStyles(self) -> None:
        """Initialize styles."""

        self.style_mapper = QtCore.QSignalMapper()
        self.style_mapper.mappedString.connect(self.setStyle)  # type: ignore
        style_read = settings.CONFIG.value("application/style", None)
        if not style_read:
            style_read = "Fusion"

        style_menu = self.mainWidget().findChild(  # type: ignore [union-attr] # noqa : F821
            QtWidgets.QMenu, "style"
        )

        if style_menu:
            from PyQt6 import QtGui

            action_group = QtGui.QActionGroup(style_menu)
            for style_ in QtWidgets.QStyleFactory.keys():
                action_ = style_menu.addAction(style_)  # type: ignore [union-attr] # noqa : F821
                if action_:
                    action_.setObjectName("style_%s" % style_)
                    action_.setCheckable(True)
                    if style_ == style_read:
                        action_.setChecked(True)

                    action_.triggered.connect(self.style_mapper.map)  # type: ignore [union-attr, arg-type]
                    self.style_mapper.setMapping(action_, style_)
                    action_group.addAction(action_)
            action_group.setExclusive(True)

        self.style = True

    def getTabWidgetPages(self, widget_name: str, obj_name: str) -> str:
        """Get tabs."""

        action_name = ""
        widget_: Optional["flformdb.FLFormDB"]
        if widget_name.startswith("formRecord"):
            action_name = widget_name[10:]
            action_ = self.db().manager().action(action_name)
            widget_ = self.db().managerModules().createFormRecord(action_)
        elif widget_name.startswith("formSearch"):
            action_name = widget_name[10:]
            action_ = self.db().manager().action(action_name)
            widget_ = self.db().managerModules().createForm(action_)
        else:
            action_name = widget_name[4:]
            action_ = self.db().manager().action(action_name)
            widget_ = self.db().managerModules().createForm(action_)

        if widget_ is None:
            return ""  # type: ignore [unreachable]

        tab_widget = cast(QtWidgets.QTabWidget, widget_.findChild(QtWidgets.QTabWidget, obj_name))
        if tab_widget is None:
            return ""  # type: ignore [unreachable]

        tab_names: str = ""
        for number in range(tab_widget.count()):
            item: Optional["QtWidgets.QWidget"] = tab_widget.widget(number)
            if item:
                tab_names += "%s/%s*" % (item.objectName(), tab_widget.tabText(number))

        return tab_names

    @decorators.not_implemented_warn
    def getWidgetList(self, widget_name: str, class_name: str) -> List:
        """Get widgets."""

        return []

    def aboutQt(self) -> None:
        """Show About QT."""
        main_widget = self.mainWidget()
        if main_widget is not None:
            QtWidgets.QMessageBox.aboutQt(main_widget)

    def aboutPineboo(self) -> None:
        """Show about Pineboo."""
        if application.PROJECT.DGI.localDesktop():
            fun_about = getattr(application.PROJECT.DGI, "about_pineboo", None)
            if fun_about is not None:
                fun_about()

    def statusHelpMsg(self, text) -> None:
        """Show help message."""

        if settings.CONFIG.value("application/isDebuggerMode", False):
            LOGGER.info("StatusHelpMsg: %s", text)

        main_widget = getattr(application.PROJECT.main_window, "main_widget", None)

        if main_widget is None:
            return

        cast(QtWidgets.QMainWindow, main_widget).statusBar().showMessage(text, 2000)  # type: ignore [union-attr]

    def loadScriptsFromModule(self, id_module: str) -> None:
        """Load scripts from named module."""
        if id_module in application.PROJECT.modules.keys():
            application.PROJECT.modules[id_module].load()

    def reinit(self) -> None:
        """Cleanup and restart."""
        if self._inicializing or self._destroying:
            return

        self.stopTimerIdle()
        # self.apAppIdle()
        self._inicializing = True

        if application.PROJECT.main_window:
            main_window = application.PROJECT.main_window

            if main_window is not None:
                main_window.writeState()
                main_window.writeStateModule()
                if hasattr(main_window, "_p_work_space"):
                    main_window._p_work_space = None

        self.reinitP()

    def startTimerIdle(self) -> None:
        """Start timer."""
        if not self.timer_idle_:
            self.timer_idle_ = QtCore.QTimer()
            self.timer_idle_.timeout.connect(  # type: ignore [attr-defined] # noqa: F821
                self.aqAppIdle
            )
        else:
            self.timer_idle_.stop()

        self.timer_idle_.start(1000)

    def stopTimerIdle(self) -> None:
        """Stop timer."""
        if self.timer_idle_ and self.timer_idle_.isActive():
            self.timer_idle_.stop()

    def aqAppIdle(self) -> None:
        """Check and fix transaction level."""
        if (
            application.PROJECT.app.activeModalWidget()
            or application.PROJECT.app.activePopupWidget()
        ):
            return

        self.checkAndFixTransactionLevel("Application::aqAppIdle()")

    def checkAndFixTransactionLevel(self, ctx=None) -> None:
        """Fix transaction."""
        dict_db = self.db().dictDatabases()
        if not dict_db:
            return

        roll_back_done = False
        for item in dict_db.values():
            if item.transactionLevel() <= 0:
                continue
            roll_back_done = True
            last_active_cursor = item.lastActiveCursor()
            if last_active_cursor is not None:
                last_active_cursor.rollbackOpened(-1)
            if item.transactionLevel() <= 0:
                continue

        if not roll_back_done:
            return

        msg = self.tr(
            "Se han detectado transacciones abiertas en estado inconsistente.\n"
            "Esto puede suceder por un error en la conexión o en la ejecución\n"
            "de algún proceso de la aplicación.\n"
            "Para mantener la consistencia de los datos se han deshecho las\n"
            "últimas operaciones sobre la base de datos.\n"
            "Los últimos datos introducidos no han sido guardados, por favor\n"
            "revise sus últimas acciones y repita las operaciones que no\n"
            "se han guardado.\n"
        )

        if ctx is not None:
            msg += self.tr("Contexto: %s\n" % ctx)

        # FIXME: Missing _gui parameter
        # self.msgBoxWarning(msg)
        LOGGER.warning("%s\n", msg)

    def clearProject(self) -> None:
        """Cleanup."""
        application.PROJECT.actions = {}
        application.PROJECT.areas = {}
        application.PROJECT.modules = {}
        # application.PROJECT.tables = {}

    def acl(self) -> Optional["pnaccesscontrollists.PNAccessControlLists"]:
        """Return acl."""
        return self.acl_

    def set_acl(self, acl: "pnaccesscontrollists.PNAccessControlLists") -> None:
        """Set acl to pineboo."""

        if application.ENABLE_ACLS:
            self.acl_ = acl
        else:
            LOGGER.warning("ACLS usage is disabled")

    def reinitP(self) -> None:
        """Reinitialize application.PROJECT."""

        from pineboolib.application.parsers.parser_mtd import pnormmodelsfactory

        self.db().managerModules().finish()
        self.db().manager().finish()
        self.setMainWidget(None)
        self.db().managerModules().setActiveIdModule("")
        self.clearProject()

        if application.PROJECT.main_window is None and not utils_base.is_library():
            from pineboolib.plugins import mainform

            main_form_name = settings.CONFIG.value("ebcomportamiento/main_form_name", "eneboo")
            main_form = getattr(mainform, main_form_name, None)
            main_form_class = getattr(main_form, "MainForm", None)
            if main_form_class is not None:
                application.PROJECT.main_window = main_form_class()
            # if application.PROJECT.main_form is not None:
            #    application.PROJECT.main_form.mainWindow = application.PROJECT.main_window.MainForm()
            #    application.PROJECT.main_window = application.PROJECT.main_form.mainWindow
            if application.PROJECT.main_window is not None:
                application.PROJECT.main_window.initScript()
                application.PROJECT.main_window.initialized_mods_ = []

        qsadictmodules.QSADictModules.clean_all()
        pnormmodelsfactory.PROCESSED = []
        application.PROJECT.files = {}
        application.PROJECT.conn_manager.useConn("default")
        application.PROJECT.conn_manager.useConn("dbaux")

        external.reload_project_config()

        application.PROJECT.run()
        # application.PROJECT.load_classes()
        # application.PROJECT.load_orm()
        # application.PROJECT.load_modules()
        self.db().managerModules().reloadStaticLoader()

        self.db().managerModules().loadIdAreas()
        self.db().managerModules().loadAllIdModules()

        self.db().manager().init()

        self.db().managerModules()
        # self.db().manager().cleanupMetaData()
        if self.acl_:
            self.acl_.init()

        self.loadScripts()
        application.PROJECT.load_orm()
        # self.db().managerModules().setShaFromGlobal()

        self._inicializing = False
        if not utils_base.is_library():
            if application.PROJECT.main_window:
                if not hasattr(application.PROJECT.main_window, "initModule"):
                    self.call("sys.init()", [])
                if hasattr(application.PROJECT.main_window, "initToolBox"):
                    application.PROJECT.main_window.initToolBox()

                # mw.readState()

                container = getattr(application.PROJECT.main_window, "container_", None)
                if container is not None:
                    container.installEventFilter(self)
                # self.container_.setDisable(False)

            self.callScriptEntryFunction()

            reinit_func = getattr(application.PROJECT.main_window, "reinitScript", None)
            if reinit_func is not None:
                reinit_func()

        LOGGER.info("Reinit completed!!")

    def showDocPage(self, url_: str) -> None:
        """Show documentation."""

        sysbasetype.SysBaseType.openUrl([url_])

    def toPixmap(self, value: str) -> "QtGui.QPixmap":
        """Create a QPixmap from a text."""

        from pineboolib.application.utils import xpm
        from PyQt6 import QtGui

        ret_ = QtGui.QPixmap()

        file_name = xpm.cache_xpm(value)
        if file_name:
            ret_ = QtGui.QPixmap(file_name)

        return ret_

    def fromPixmap(self, pix_: "QtGui.QPixmap") -> str:
        """Return a text from a QPixmap."""
        ret_: str = ""
        if pix_.isNull():
            return ret_

        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.WaitCursor)
        buffer_ = QtCore.QBuffer()
        buffer_.open(QtCore.QIODevice.OpenModeFlag.WriteOnly)
        pix_.save(buffer_, "xpm")

        application.PROJECT.app.restoreOverrideCursor()

        return str(buffer_.data())

    def scalePixmap(
        self, pix_: "QtGui.QPixmap", width: int, height: int, mode_: "QtCore.Qt.AspectRatioMode"
    ) -> "QtGui.QImage":
        """Return QImage scaled from a QPixmap."""

        img_ = pix_.toImage()

        return img_.scaled(height, height, mode_)

    def timeUser(self) -> "str":
        """Get amount of time running."""

        return QtCore.QDateTime.currentDateTime().toString(QtCore.Qt.DateFormat.ISODate)

    def call(self, function, argument_list=[], object_content=None, show_exceptions=True) -> Any:
        """Call a QS project function."""
        return application.PROJECT.call(function, argument_list, object_content, show_exceptions)

    @decorators.not_implemented_warn
    def setNotExit(self, value: bool) -> None:
        """Protect against window close."""
        self._not_exit = value

    def setNotifyEndTransaction(self, value: bool) -> None:
        """Set notify end transaction mode."""

        DB_SIGNALS.notify_end_transaction_ = value

    def setNotifyBeginTransaction(self, value: bool) -> None:
        """Set notify begin transaction mode."""

        DB_SIGNALS.notify_begin_transaction_ = value

    def setNotifyRollbackTransaction(self, value: bool) -> None:
        """Set notify rollback transaction mode."""

        DB_SIGNALS.notify_roll_back_transaction_ = value

    def notifyBeginTransaction(self) -> bool:
        """Return if notify begin transaction is enabled."""

        return DB_SIGNALS.notify_begin_transaction_

    def notifyEndTransaction(self) -> bool:
        """Return if notify end transaction is enabled."""

        return DB_SIGNALS.notify_end_transaction_

    def notifyRollbackTransaction(self) -> bool:
        """Return if notify rollback transaction is enabled."""

        return DB_SIGNALS.notify_roll_back_transaction_

    @decorators.not_implemented_warn
    def printTextEdit(self, editor_):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def setPrintProgram(self, print_program_):
        """Not implemented."""
        pass

    def setCaptionMainWidget(self, text: str) -> None:
        """Set caption main widget."""
        if application.PROJECT.main_window is not None:
            application.PROJECT.main_window.setCaptionMainWidget(text)

    @decorators.not_implemented_warn
    def addSysCode(self, code, script_entry_function):
        """Not implemented."""
        pass

    def setScriptEntryFunction(self, script_enttry_function) -> None:
        """Set which QS function to call on startup."""
        self.script_entry_function_ = script_enttry_function

    @decorators.not_implemented_warn
    def setDatabaseLockDetection(
        self, on_, msec_lapsus, lim_checks, show_warn, msg_warn, connection_name
    ):
        """Not implemented."""
        pass

    def popupWarn(self, msg_warn: str, script_call: List[Any] = []) -> None:
        """Show a warning popup."""
        main_window = application.PROJECT.main_window

        if main_window is None:
            return

        if script_call:
            self.call(script_call, [], self)

        if not main_window.isHidden():
            QtWidgets.QWhatsThis.showText(
                main_window.mapToGlobal(QtCore.QPoint(main_window.width() * 2, 0)),
                msg_warn,
                main_window,
            )
            QtCore.QTimer.singleShot(2000, QtWidgets.QWhatsThis.hideText)
            application.PROJECT.app.processEvents()  # type: ignore[misc] # noqa: F821

    @decorators.not_implemented_warn
    def checkDatabaseLocks(self, timer_):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def saveGeometryForm(self, name, geo):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def geometryForm(self, name):
        """Not implemented."""
        pass

    def staticLoaderSetup(self) -> None:
        """Initialize static loader."""
        self.db().managerModules().staticLoaderSetup()

    def mrProper(self) -> None:
        """Cleanup database."""
        self.db().mainConn().Mr_Proper()

    def showConsole(self) -> None:
        """Show application console on GUI."""

        if application.PROJECT.main_window is not None:
            if self._ted_output:
                self._ted_output.parentWidget().close()  # type: ignore [union-attr]

            dock_widget = QtWidgets.QDockWidget("tedOutputDock", application.PROJECT.main_window)

            if dock_widget is not None:
                self._ted_output = TextEditOutput(dock_widget)
                dock_widget.setWidget(self._ted_output)
                dock_widget.setWindowTitle(self.tr("Mensajes de Eneboo"))
                application.PROJECT.main_window.addDockWidget(
                    QtCore.Qt.DockWidgetArea.BottomDockWidgetArea, dock_widget
                )

    def consoleShown(self) -> bool:
        """Return if console is shown."""

        hidden = getattr(self._ted_output, "isHidden", None)
        return False if hidden is None else hidden()

    def modMainWidget(self, id_modulo: str) -> Optional["QtWidgets.QWidget"]:
        """Set module main widget."""

        main_window = application.PROJECT.main_window
        if main_window is None:
            return None

        mod_widget: Optional[QtWidgets.QWidget] = None
        dict_main_widgets = getattr(main_window, "_dict_main_widgets", {})
        if id_modulo in dict_main_widgets.keys():
            mod_widget = dict_main_widgets[id_modulo]

        if mod_widget is None:
            list_ = application.PROJECT.app.topLevelWidgets()
            for widget in list_:
                if widget.objectName() == id_modulo:
                    mod_widget = widget
                    break

        if mod_widget is None and self.mainWidget() is not None:
            mod_widget = cast(
                QtWidgets.QWidget, main_window.main_widget.findChild(QtWidgets.QWidget, id_modulo)
            )

        return mod_widget

    def evaluateProject(self) -> None:
        """Execute QS entry function."""
        QtCore.QTimer.singleShot(0, self.callScriptEntryFunction)

    def callScriptEntryFunction(self) -> None:
        """Execute QS entry function."""
        if self.script_entry_function_:
            self.call(self.script_entry_function_, [], self)
            # self.script_entry_function_ = None

    def emitTransactionBegin(
        self, cursor: Union["pnsqlcursor.PNSqlCursor", "isqlcursor.ISqlCursor"]
    ) -> None:
        """Emit signal."""
        self.transactionBegin.emit()
        DB_SIGNALS.emitTransactionBegin(cursor)

    def emitTransactionEnd(
        self, cursor: Union["pnsqlcursor.PNSqlCursor", "isqlcursor.ISqlCursor"]
    ) -> None:
        """Emit signal."""
        self.transactionEnd.emit()
        DB_SIGNALS.emitTransactionEnd(cursor)

    def emitTransactionRollback(
        self, cursor: Union["pnsqlcursor.PNSqlCursor", "isqlcursor.ISqlCursor"]
    ) -> None:
        """Emit signal."""
        self.transactionRollback.emit()
        DB_SIGNALS.emitTransactionRollback(cursor)

    def self_(self) -> "PNApplication":
        """Return self."""

        return self

    @decorators.not_implemented_warn
    def gsExecutable(self):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def evalueateProject(self):
        """Not implemented."""
        pass

    def DGI(self) -> "dgi_schema.dgi_schema":
        """Return current DGI."""
        return application.PROJECT.DGI

    def singleFLLarge(self) -> bool:
        """
        Para especificar si usa fllarge unificado o multiple (Eneboo/Abanq).

        @return True (Tabla única), False (Múltiples tablas)
        """
        if self.flLargeMode is None:
            ret = utils.sql_select("flsettings", "valor", "flkey='FLLargeMode'")
            self.flLargeMode = False if ret in ["True", True] else True
        return self.flLargeMode

    def msgBoxWarning(self, text: str, _gui: Any) -> None:
        """Display warning."""
        _gui.msgBoxWarning(text)

    @decorators.not_implemented_warn
    def showDebug(self):
        """Return if debug is shown."""
        return self.show_debug_

    def db(self) -> "pnconnectionmanager.PNConnectionManager":
        """Return current connection."""
        return application.PROJECT.conn_manager

    def classType(self, obj) -> str:
        """Return class for object."""

        return str(type(obj))

    def mainWidget(self) -> Optional["QtWidgets.QWidget"]:
        """Return current mainWidget."""
        return getattr(application.PROJECT.main_window, "main_widget", None)

    def quit(self) -> None:
        """Handle quit/close signal."""
        main_window = application.PROJECT.main_window
        if main_window is not None:
            main_window.close()

    def queryExit(self) -> bool:
        """Ask user if really wants to quit."""

        if self._not_exit:
            return False

        if application.PROJECT.conn_manager.mainConn().interactiveGUI():
            main_widget = application.PROJECT.main_window
            if main_widget is not None:
                ret = QtWidgets.QMessageBox.question(
                    main_widget,
                    self.tr("Salir ..."),
                    self.tr("¿ Quiere salir de la aplicación ?"),
                    QtWidgets.QMessageBox.StandardButton.Yes,
                    QtWidgets.QMessageBox.StandardButton.No,
                )
                return ret == QtWidgets.QMessageBox.StandardButton.Yes

        return True

    def loadScripts(self) -> None:
        """Load scripts for all modules."""

        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.WaitCursor)
        list_modules = self.db().managerModules().listAllIdModules()
        for item in list_modules:
            self.loadScriptsFromModule(item)

        application.PROJECT.app.restoreOverrideCursor()

    def urlPineboo(self) -> None:
        """Open Eneboo URI."""
        sysbasetype.SysBaseType.openUrl(["http://eneboo.org/"])

    def helpIndex(self) -> None:
        """Open help."""
        sysbasetype.SysBaseType.openUrl(["http://manuales-eneboo-pineboo.org/"])

    def loadTranslations(self) -> None:
        """
        Install loaded translations.
        """

        lang = QtCore.QLocale().name()[:2]
        lang = "es" if lang == "C" else lang

        for module in self.modules().keys():
            self.loadTranslationFromModule(module, lang)

        for item in self._translator:
            self.removeTranslator(item)
            if item._sys_trans:
                self.installTranslator(item)

    def trMulti(self, text: str, lang: str):
        """
        Lookup translation for certain language.

        @param text, Cadena de texto
        @param lang, Idioma.
        @return Cadena de texto traducida.
        """
        return application.PROJECT.app.tr("%s_MULTILANG" % lang.upper(), text)

    def setMultiLang(self, enable_: bool, lang_id_: str) -> None:
        """
        Change multilang status.

        @param enable, Boolean con el nuevo estado
        @param langid, Identificador del leguaje a activar
        """
        self._multi_lang_enabled = enable_
        if enable_ and lang_id_:
            self._multi_lang_id = lang_id_.upper()

    def loadTranslationFromModule(self, id_module: str, lang: str) -> None:
        """
        Load translation from module.

        @param idM, Identificador del módulo donde buscar
        @param lang, Lenguaje a buscar
        """
        self.installTranslator(self.createModTranslator(id_module, lang, True))

    def installTranslator(self, tor) -> None:
        """
        Install translation for app.

        @param tor, Objeto con la traducción a cargar
        """

        if tor is not None:
            application.PROJECT.app.installTranslator(tor)
            self._translator.append(tor)

    def removeTranslator(self, tor) -> None:
        """
        Delete translation on app.

        @param tor, Objeto con la traducción a cargar
        """
        if tor is None:
            return
        else:
            application.PROJECT.app.removeTranslator(tor)
            if tor in self._translator:
                self._translator.remove(tor)

    @decorators.not_implemented_warn
    def createSysTranslator(self, lang, load_default):
        """
        Create SYS Module translation.

        @param lang, Idioma a usar
        @param loadDefault, Boolean para cargar los datos por defecto
        @return objeto traducción
        """
        pass

    def createModTranslator(
        self, id_module, lang: str, load_default: bool = False
    ) -> Optional["pntranslator.PNTranslator"]:
        """
        Create new translation for module.

        @param idM, Identificador del módulo
        @param lang, Idioma a usar
        @param loadDefault, Boolean para cargar los datos por defecto
        @return objeto traducción
        """
        file_ts = "%s.%s.ts" % (id_module, lang)
        key = None

        if id_module == "sys":
            key = " "

        else:
            if self.db():
                key = self.db().managerModules().shaOfFile(file_ts)

        if key:
            from pineboolib.application.translator import pntranslator

            tor = pntranslator.PNTranslator(
                self.mainWidget(), "%s_%s" % (id_module, lang), lang == "multilang"
            )
            if key and tor.loadTsContent(key):
                return tor

        return self.createModTranslator(id_module, "es") if load_default else None

    def modules(self) -> Dict[str, "app_module.Module"]:
        """Return loaded modules."""
        return application.PROJECT.modules

    def commaSeparator(self) -> str:
        """Return comma separator for floating points on current language."""
        return self.comma_separator

    def tmp_dir(self) -> str:
        """Return temporary folder."""
        return application.PROJECT.tmpdir

    def applicationDirPath(self) -> str:
        """Return application dir path."""
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
        # return application.PROJECT.app.applicationDirPath()

    def transactionLevel(self):
        """Return number of concurrent transactions."""
        return application.PROJECT.conn_manager.useConn("default").transactionLevel()

    def version(self):
        """Return app version."""
        return application.PROJECT.load_version()

    def dialogGetFileImage(self) -> Optional[str]:
        """Get image file name."""

        file_dialog = QtWidgets.QFileDialog(
            QtWidgets.QApplication.focusWidget(),
            self.tr("Elegir archivo"),
            application.PROJECT.tmpdir,
            "*",
        )
        # pixmap_viewer = flpixmapview.FLPixmapView(file_dialog)

        # pixmap_viewer.setAutoScaled(True)
        # file_dialog.setContentsPreviewEnabled(True)
        # file_dialog.setContentsPreview(p, p)
        # file_dialog.setPreviewMode(QtWidgets.QFileDialog.Contents)

        file_name = None
        if file_dialog.exec() == QtWidgets.QDialog.DialogCode.Accepted:
            list_ = file_dialog.selectedFiles()
            if list_:
                file_name = list_[0]

        return file_name


class TextEditOutput(QtWidgets.QPlainTextEdit):
    """FLTextEditOutput class."""

    oldStdout: TextIO
    oldStderr: TextIO

    def __init__(self, parent: "QtWidgets.QWidget") -> None:
        """Inicialize."""
        super().__init__(parent)

        self.oldStdout = sys.stdout  # pylint: disable=invalid-name
        self.oldStderr = sys.stderr  # pylint: disable=invalid-name
        sys.stdout = self  # type: ignore [assignment] # noqa F821
        sys.stderr = self  # type: ignore [assignment] # noqa F821
        self.setReadOnly(True)

    def write(self, txt: Union[bytearray, bytes, str]) -> None:
        """Set text."""
        txt = str(txt)
        if self.oldStdout:
            self.oldStdout.write(txt)
        self.appendPlainText(txt)

    def flush(self):
        """Flush data."""

        pass

    def close(self) -> bool:
        """Control close."""
        if self.oldStdout:
            sys.stdout = self.oldStdout
        if self.oldStderr:
            sys.stderr = self.oldStderr
        return super().close()


# aqApp = FLApplication()
aqApp: "PNApplication"



"""Flnetwork module."""

# # -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtNetwork  # type: ignore[import]

from typing import Optional, cast, Any


from pineboolib.core import decorators


class FLNetwork(QtCore.QObject):
    """FLNetwork class."""

    url: str
    request: QtNetwork.QNetworkRequest
    manager: QtNetwork.QNetworkAccessManager

    reply: Optional[QtNetwork.QNetworkReply] = None

    finished = QtCore.pyqtSignal()
    start = QtCore.pyqtSignal()
    data = QtCore.pyqtSignal(str)
    dataTransferProgress = QtCore.pyqtSignal(int, int)

    def __init__(self, url: str) -> None:
        """Inicialize."""

        super(FLNetwork, self).__init__()
        self.url = url

        self.request = QtNetwork.QNetworkRequest()

        self.manager = QtNetwork.QNetworkAccessManager()
        # self.manager.readyRead.connect(self._slotNetworkStart)
        cast(
            QtCore.pyqtSignal, self.manager.finished
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self._slotNetworkFinished
        )
        # finished_signal["QNetworkReply*"].connect(self._slotNetworkFinished) # FIXME: What does this code?
        # self.data.connect(self._slotNetWorkData)
        # self.dataTransferProgress.connect(self._slotNetworkProgress)

    @decorators.beta_implementation
    def get(self, location: str) -> None:
        """Get value from a location."""

        self.request.setUrl(QtCore.QUrl("%s%s" % (self.url, location)))
        self.reply = cast(QtNetwork.QNetworkReply, self.manager.get(self.request))
        try:
            cast(
                QtCore.pyqtSignal, self.reply.uploadProgress
            ).disconnect(  # type: ignore [attr-defined] # noqa: F821
                self._slotNetworkProgress
            )
            cast(
                QtCore.pyqtSignal, self.reply.downloadProgress
            ).disconnect(  # type: ignore [attr-defined] # noqa: F821
                self._slotNetworkProgress
            )
        except Exception:
            pass

        cast(
            QtCore.pyqtSignal, self.reply.downloadProgress
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self._slotNetworkProgress
        )

    @decorators.beta_implementation
    def put(self, data: Any, location: str) -> None:
        """Send data to a location."""

        self.request.setUrl(QtCore.QUrl("%s%s" % (self.url, location)))
        self.reply = cast(QtNetwork.QNetworkReply, self.manager.put(self.request, data))
        try:
            cast(
                QtCore.pyqtSignal, self.reply.uploadProgress
            ).disconnect(  # type: ignore [attr-defined] # noqa: F821
                self._slotNetworkProgress
            )
            cast(
                QtCore.pyqtSignal, self.reply.downloadProgress
            ).disconnect(  # type: ignore [attr-defined] # noqa: F821
                self._slotNetworkProgress
            )
        except Exception:
            pass
        cast(
            QtCore.pyqtSignal, self.reply.uploadProgress
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self._slotNetworkProgress
        )

    @decorators.beta_implementation
    def copy(self, from_location: str, to_location: str) -> None:
        """Copy data from a location to another."""

        self.request.setUrl(QtCore.QUrl("%s%s" % (self.url, from_location)))
        data = self.manager.get(self.request)
        self.put(data.readAll(), to_location)  # type: ignore [union-attr]

    @decorators.pyqt_slot()
    def _slotNetworkStart(self) -> None:
        """Emit start signal."""

        self.start.emit()

    @decorators.pyqt_slot()
    def _slotNetworkFinished(self, reply: Any = None) -> None:
        """Emit finished signal."""

        self.finished.emit()

    # @decorators.pyqt_slot(QtCore.QByteArray)
    # def _slotNetWorkData(self, b):
    #    buffer = b
    #    self.data.emit(b)

    def _slotNetworkProgress(self, bytes_done: int, bytes_total: int) -> None:
        """Process data received."""

        if self.reply is None:
            raise Exception("No reply in progress")
        self.dataTransferProgress.emit(bytes_done, bytes_total)
        data_ = None
        reply_ = self.reply.readAll().data()
        try:
            data_ = str(reply_, encoding="iso-8859-15")
        except Exception:
            data_ = str(reply_, encoding="utf-8")

        self.data.emit(data_)



"""Aqssproject module."""

from PyQt6 import QtCore  # type: ignore[import]

from pineboolib import application


class AQSSProject(QtCore.QObject):
    """AQSSProject class."""

    New = 0
    Changed = 1
    UnChanged = 2

    def callEntryFunction(self):
        """Call entry function."""
        application.PROJECT.aq_app.callScriptEntryFunction()

    def get_entry_function(self) -> str:
        """Return entry function."""
        return application.PROJECT.aq_app.script_entry_function_

    def set_entry_function(self, entry_fun_: str) -> None:
        """Set entry function."""
        application.PROJECT.aq_app.script_entry_function_ = entry_fun_

    entryFunction = property(get_entry_function, set_entry_function)



"""Test_pnbuffer module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.database import pnsqlcursor, pnsqlquery


class TestPNBuffer(unittest.TestCase):
    """TestPNBuffer Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Basic test."""
        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "Campo texto 1")
        cursor.setValueBuffer("date_field", "2019-01-01")
        cursor.setValueBuffer("time_field", "01:01:01")
        cursor.setValueBuffer("bool_field", False)
        cursor.setValueBuffer("double_field", 1.01)
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "Campo texto 2")
        cursor.setValueBuffer("date_field", "2019-02-02T00:00:00")
        cursor.setValueBuffer("time_field", "02:02:02.1234")
        cursor.setValueBuffer("bool_field", "true")
        cursor.setValueBuffer("double_field", 2.02)
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "Campo texto 3")
        cursor.setValueBuffer("date_field", "2019-03-03")
        cursor.setValueBuffer("time_field", "2019-03-03T03:03:03 +2")
        cursor.setValueBuffer("bool_field", "false")
        cursor.setValueBuffer("double_field", 3.03)
        self.assertTrue(cursor.commitBuffer())

    def test_basic1(self) -> None:
        """Basic test 1."""

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.select()
        self.assertEqual(cursor.size(), 3)
        self.assertTrue(cursor.first())

        buffer_ = cursor.buffer()
        self.assertTrue(buffer_ is not None)
        if buffer_ is not None:
            buffer_.prime_update()

            self.assertEqual(buffer_.value("date_field"), "2019-01-01T00:00:00")
            self.assertEqual(buffer_.value("time_field"), "01:01:01")

        self.assertEqual(cursor.valueBuffer("date_field"), "2019-01-01T00:00:00")

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("date_field, time_field")
        qry.setFrom("fltest")
        qry.setWhere("1 = 1")
        qry.setOrderBy("string_field ASC")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(qry.value(0), "2019-01-01T00:00:00")
        self.assertEqual(qry.value(1), "01:01:01")
        self.assertTrue(qry.next())
        self.assertEqual(qry.value(1), "02:02:02")
        self.assertTrue(qry.next())
        self.assertEqual(qry.value(1), "03:03:03")
        self.assertEqual(qry.value(0), "2019-03-03T00:00:00")

    def test_basic2(self) -> None:
        """Basic test 2."""

        # import datetime

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.select()
        cursor.first()

        buffer_ = cursor.buffer()
        self.assertTrue(buffer_ is not None)
        if buffer_ is not None:
            self.assertEqual(buffer_.is_null("string_field"), False)
            self.assertEqual(buffer_.is_null("date_field"), False)
            self.assertEqual(buffer_.is_null("time_field"), False)
            self.assertEqual(buffer_.is_null("bool_field"), False)
            self.assertEqual(buffer_.is_null("double_field"), False)

            buffer_.prime_update()

            self.assertEqual(buffer_.value("string_field"), "Campo texto 1")
            self.assertEqual(buffer_.value("double_field"), 1.01)
            self.assertEqual(buffer_.value("date_field"), "2019-01-01T00:00:00")
            self.assertEqual(buffer_.value("time_field"), "01:01:01")
            self.assertEqual(buffer_.value("bool_field"), False)

    def test_basic3(self) -> None:
        """Basic test 3."""

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.select()
        cursor.first()

        buffer_ = cursor.buffer()
        self.assertTrue(buffer_ is not None)

        if buffer_ is not None:
            buffer_.set_value("string_field", "Campo texto 1 mod")
            buffer_.set_value("double_field", 1.02)
            self.assertEqual(buffer_.value("double_field"), 1.02)

    def test_null(self) -> None:
        """Test null."""

        cursor_1 = pnsqlcursor.PNSqlCursor("fltest")
        cursor_1.setModeAccess(cursor_1.Insert)
        cursor_1.refreshBuffer()
        self.assertFalse(cursor_1.isNull("id"))
        self.assertEqual(cursor_1.buffer().value("uint_field"), 0)
        self.assertEqual(cursor_1.buffer().value("uint_field", True), None)
        self.assertTrue(cursor_1.isNull("string_field"))
        self.assertFalse(cursor_1.isNull("double_field"))  # default 0

        cursor_1.setNull("double_field")
        self.assertEqual(cursor_1.buffer().value("double_field"), 0)
        self.assertEqual(cursor_1.buffer().value("double_field", True), None)
        self.assertTrue(cursor_1.isNull("double_field"))

        self.assertEqual(cursor_1.valueBuffer("uint_field"), 0)
        self.assertEqual(cursor_1.buffer().value("uint_field"), 0)
        self.assertEqual(cursor_1.buffer().value("uint_field", True), None)

        cursor_2 = pnsqlcursor.PNSqlCursor("fltest5")
        cursor_2.setModeAccess(cursor_2.Insert)
        cursor_2.refreshBuffer()
        self.assertTrue(cursor_2.isNull("unit_field"))
        self.assertTrue(cursor_2.isNull("uint_field"))
        self.assertEqual(cursor_2.valueBuffer("uint_field"), 0)
        self.assertEqual(cursor_2.buffer().value("uint_field", True), None)
        self.assertEqual(cursor_2.buffer().value("uint_field"), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()
